from typing import Sequence

import numpy as np
from scipy.stats import spearmanr

from pydvl.utils import Utility
from pydvl.value import ValuationResult


def polynomial(coefficients, x):
    powers = np.arange(len(coefficients))
    return np.power(x, np.tile(powers, (len(x), 1)).T).T @ coefficients


def check_total_value(
    u: Utility, values: ValuationResult, rtol: float = 0.05, atol: float = 1e-6
):
    """Checks absolute distance between total and added values.
    Shapley value is supposed to fulfill the total value axiom."""
    total_utility = u(u.data.indices)
    # We can use relative tolerances if we don't have the range of the scorer.
    assert np.isclose(np.sum(values.values), total_utility, rtol=rtol, atol=atol)


def check_exact(
    values: ValuationResult,
    exact_values: ValuationResult,
    rtol: float = 0.1,
    atol: float = 1e-6,
):
    """Compares ranks and values."""

    values.sort()
    exact_values.sort()

    assert np.all(values.indices == exact_values.indices), "Ranks do not match"
    assert np.allclose(
        values.values, exact_values.values, rtol=rtol, atol=atol
    ), "Values do not match"


def check_values(
    values: ValuationResult,
    exact_values: ValuationResult,
    rtol: float = 0.1,
    atol: float = 1e-5,
    *,
    extra_values_names: Sequence[str] = tuple(),
):
    """Compares values in dictionaries.

    Asserts that `|value - exact_value| < |exact_value| * rtol + atol` for
    all pairs of `value`, `exact_value` with equal keys.

    Note that this does not assume any ordering (despite values typically being
    stored in an OrderedDict elsewhere.

    :param values:
    :param exact_values:
    :param rtol: relative tolerance of elements in `values` with respect to
        elements in `exact_values`. E.g. if rtol = 0.1, and atol = 0 we must
        have |value - exact_value|/|exact_value| < 0.1 for every value
    :param atol: absolute tolerance of elements in `values` with respect to
        elements in `exact_values`. E.g. if atol = 0.1, and rtol = 0 we must
        have |value - exact_value| < 0.1 for every value.
    :param extra_values_names: Sequence of names of extra values that should
        also be compared.
    """
    values.sort()
    exact_values.sort()

    assert np.allclose(values.values, exact_values.values, rtol=rtol, atol=atol)
    for name in extra_values_names:
        assert np.isclose(
            getattr(values, name), getattr(exact_values, name), rtol=rtol, atol=atol
        )


def check_rank_correlation(
    values: ValuationResult,
    exact_values: ValuationResult,
    k: int = None,
    threshold: float = 0.9,
):
    """Checks that the indices of `values` and `exact_values` follow the same
    order (by value), with some slack, using Spearman's correlation.

    Runs an assertion for testing.

    :param values: The values and indices to test
    :param exact_values: The ground truth
    :param k: Consider only these many, starting from the top.
    :param threshold: minimal value for spearman correlation for the test to
        succeed
    """
    # FIXME: estimate proper threshold for spearman

    k = k or len(values)

    values.sort()
    exact_values.sort()

    top_k = np.array([it.index for it in values[-k:]])
    top_k_exact = np.array([it.index for it in exact_values[-k:]])

    correlation, pvalue = spearmanr(top_k, top_k_exact)
    assert correlation >= threshold, f"{correlation} < {threshold}"



import operator
from enum import Enum
from functools import reduce
from itertools import cycle, takewhile
from typing import Sequence

import numpy as np
from numpy.typing import NDArray
from tqdm import tqdm

from pydvl.utils import MapReduceJob, ParallelConfig, Utility, random_powerset
from pydvl.value import ValuationResult
from pydvl.value.stopping import MinUpdates

__all__ = ["OwenAlgorithm", "owen_sampling_shapley"]


class OwenAlgorithm(Enum):
    Standard = "standard"
    Antithetic = "antithetic"


def _owen_sampling_shapley(
    indices: Sequence[int],
    u: Utility,
    method: OwenAlgorithm,
    n_samples: int,
    max_q: int,
    *,
    progress: bool = False,
    job_id: int = 1,
) -> ValuationResult:
    r"""This is the algorithm as detailed in the paper: to compute the outer
    integral over q ∈ [0,1], use uniformly distributed points for evaluation
    of the integrand. For the integrand (the expected marginal utility over the
    power set), use Monte Carlo.

    .. todo::
        We might want to try better quadrature rules like Gauss or Rombert or
        use Monte Carlo for the double integral.

    :param indices: Indices to compute the value for
    :param u: Utility object with model, data, and scoring function
    :param method: Either :attr:`~OwenAlgorithm.Full` for $q \in [0,1]$ or
        :attr:`~OwenAlgorithm.Halved` for $q \in [0,0.5]$ and correlated samples
    :param n_samples: Number of subsets to sample to estimate the integrand
    :param max_q: number of subdivisions for the integration over $q$
    :param progress: Whether to display progress bars for each job
    :param job_id: For positioning of the progress bar
    :return: Object with the data values, errors.
    """
    q_stop = {OwenAlgorithm.Standard: 1.0, OwenAlgorithm.Antithetic: 0.5}
    q_steps = np.linspace(start=0, stop=q_stop[method], num=max_q)

    result = ValuationResult.zeros(
        algorithm="owen_sampling_shapley_" + str(method),
        indices=np.array(indices, dtype=np.int_),
        data_names=[u.data.data_names[i] for i in indices],
    )

    done = MinUpdates(1)
    repeat_indices = takewhile(lambda _: not done(result), cycle(indices))
    pbar = tqdm(disable=not progress, position=job_id, total=100, unit="%")
    for idx in repeat_indices:
        pbar.n = 100 * done.completion()
        pbar.refresh()
        e = np.zeros(max_q)
        subset = np.setxor1d(u.data.indices, [idx], assume_unique=True)
        for j, q in enumerate(q_steps):
            for s in random_powerset(subset, n_samples=n_samples, q=q):
                marginal = u({idx}.union(s)) - u(s)
                if method == OwenAlgorithm.Antithetic and q != 0.5:
                    s_complement = np.setxor1d(subset, s, assume_unique=True)
                    marginal += u({idx}.union(s_complement)) - u(s_complement)
                    marginal /= 2
                e[j] += marginal
        e /= n_samples
        result.update(idx, e.mean())
        # Trapezoidal rule
        # TODO: investigate whether this or other quadrature rules are better
        #  than a simple average
        # result.update(idx, (e[:-1] + e[1:]).sum() / (2 * max_q))

    return result


def owen_sampling_shapley(
    u: Utility,
    n_samples: int,
    max_q: int,
    *,
    method: OwenAlgorithm = OwenAlgorithm.Standard,
    n_jobs: int = 1,
    config: ParallelConfig = ParallelConfig(),
    progress: bool = False,
) -> ValuationResult:
    r"""Owen sampling of Shapley values as described in
    :footcite:t:`okhrati_multilinear_2021`.

    This function computes a Monte Carlo approximation to

    $$v_u(i) = \int_0^1 \mathbb{E}_{S \sim P_q(D_{\backslash \{i\}})}
    [u(S \cup \{i\}) - u(S)]$$

    using one of two methods. The first one, selected with the argument ``mode =
    OwenAlgorithm.Standard``, approximates the integral with:

    $$\hat{v}_u(i) = \frac{1}{Q M} \sum_{j=0}^Q \sum_{m=1}^M [u(S^{(q_j)}_m
    \cup \{i\}) - u(S^{(q_j)}_m)],$$

    where $q_j = \frac{j}{Q} \in [0,1]$ and the sets $S^{(q_j)}$ are such that a
    sample $x \in S^{(q_j)}$ if a draw from a $Ber(q_j)$ distribution is 1.

    The second method, selected with the argument ``mode =
    OwenAlgorithm.Antithetic``, uses correlated samples in the inner sum to
    reduce the variance:

    $$\hat{v}_u(i) = \frac{1}{2 Q M} \sum_{j=0}^Q \sum_{m=1}^M [u(S^{(q_j)}_m
    \cup \{i\}) - u(S^{(q_j)}_m) + u((S^{(q_j)}_m)^c \cup \{i\}) - u((S^{(
    q_j)}_m)^c)],$$

    where now $q_j = \frac{j}{2Q} \in [0,\frac{1}{2}]$, and $S^c$ is the
    complement of $S$.

    .. note::
       The outer integration could be done instead with a quadrature rule.

    :param u: :class:`~pydvl.utils.utility.Utility` object holding data, model
        and scoring function.
    :param n_samples: Numer of sets to sample for each value of q
    :param max_q: Number of subdivisions for q ∈ [0,1] (the element sampling
        probability) used to approximate the outer integral.
    :param method: Selects the algorithm to use, see the description. Either
        :attr:`~OwenAlgorithm.Full` for $q \in [0,1]$ or
        :attr:`~OwenAlgorithm.Halved` for $q \in [0,0.5]$ and correlated samples
    :param n_jobs: Number of parallel jobs to use. Each worker receives a chunk
        of the total of `max_q` values for q.
    :param config: Object configuring parallel computation, with cluster
        address, number of cpus, etc.
    :param progress: Whether to display progress bars for each job.
    :return: Object with the data values.

    .. versionadded:: 0.3.0

    .. versionchanged:: 0.5.0
       Support for parallel computation and enable antithetic sampling.

    """
    map_reduce_job: MapReduceJob[NDArray, ValuationResult] = MapReduceJob(
        u.data.indices,
        map_func=_owen_sampling_shapley,
        reduce_func=lambda results: reduce(operator.add, results),
        map_kwargs=dict(
            u=u,
            method=OwenAlgorithm(method),
            n_samples=n_samples,
            max_q=max_q,
            progress=progress,
        ),
        n_jobs=n_jobs,
        config=config,
    )

    return map_reduce_job()



import logging

import numpy as np
from sklearn import datasets
from sklearn.metrics import make_scorer
from sklearn.neighbors import KNeighborsClassifier

from pydvl.utils.dataset import Dataset
from pydvl.utils.parallel.backend import available_cpus
from pydvl.utils.score import Scorer
from pydvl.utils.utility import Utility
from pydvl.value.shapley.knn import knn_shapley
from pydvl.value.shapley.naive import combinatorial_exact_shapley

log = logging.getLogger(__name__)


def test_knn_montecarlo_match(seed):
    data = Dataset.from_sklearn(
        datasets.load_iris(),
        train_size=0.05,
        random_state=seed,
        stratify_by_target=True,
    )
    model = KNeighborsClassifier(n_neighbors=5)
    u = Utility(model=model, data=data)
    knn_values = knn_shapley(u, progress=False)

    def knn_loss_function(labels, predictions, n_classes=3):
        log.debug(f"{predictions=}")
        if len(predictions[0]) < n_classes:
            raise RuntimeError("Found less classes than expected.")
        pred_proba = [predictions[i][label] for i, label in enumerate(labels)]
        return np.mean(pred_proba)

    scorer = Scorer(
        make_scorer(knn_loss_function, greater_is_better=True, needs_proba=True),
        default=0.0,
        range=(0.0, 1.0),
    )

    utility = Utility(
        model, data=data, scorer=scorer, show_warnings=False, enable_cache=False
    )
    exact_values = combinatorial_exact_shapley(
        utility, progress=False, n_jobs=min(len(data), available_cpus())
    )

    # will check only matching top elements since the scoring functions are not
    # exactly the same
    knn_values.sort()
    exact_values.sort()
    top_knn = knn_values.indices[-2:]
    top_exact = exact_values.indices[-4:]
    assert np.all([k in top_exact for k in top_knn])



from abc import ABC
from typing import Callable, Iterable, Optional, Tuple

from numpy import ndarray

__all__ = [
    "TwiceDifferentiable",
    "MatrixVectorProduct",
    "MatrixVectorProductInversionAlgorithm",
]


class TwiceDifferentiable(ABC):
    def num_params(self) -> int:
        pass

    def split_grad(self, x: ndarray, y: ndarray, progress: bool = False) -> ndarray:
        """
        Calculate the gradient of the model wrt each input x and labels y.
        The output is therefore of size [Nxp], with N the amout of points (the length of x and y) and
        P the number of parameters.
        """
        pass

    def grad(self, x: ndarray, y: ndarray) -> Tuple[ndarray, ndarray]:
        """
        It calculates the gradient of model parameters with respect to input x and labels y.
        """
        pass

    def mvp(
        self,
        grad_xy: ndarray,
        v: ndarray,
        progress: bool = False,
        backprop_on: Optional[Iterable] = None,
    ) -> ndarray:
        """
        Calculate the hessian vector product over the loss with all input parameters x and y with the vector v.
        """
        pass


MatrixVectorProduct = Callable[[ndarray], ndarray]

MatrixVectorProductInversionAlgorithm = Callable[
    [MatrixVectorProduct, ndarray], ndarray
]



# -*- coding: utf-8 -*-
#
# pyDVL documentation build configuration file
#
# This file is execfile()d with the current directory set to its containing dir.
#
# All configuration values have a default; values that are commented out
# serve to show the default.

import ast
import logging
import os
import sys
from pathlib import Path

import pkg_resources

logger = logging.getLogger("docs")

ROOT_DIR = Path(__file__).resolve().parent.parent

# If extensions (or modules to document with autodoc) are in another directory,
# add these directories to sys.path here. If the directory is relative to the
# documentation root, use os.path.abspath to make it absolute, like shown here.
sys.path.insert(0, os.fspath(ROOT_DIR / "src"))

# For custom extensions
sys.path.append(os.path.abspath("_ext"))

# -- General configuration -----------------------------------------------------

# If your documentation needs a minimal Sphinx version, state it here.
# needs_sphinx = '1.0'

# Add any Sphinx extension module names here, as strings. They can be extensions
# coming with Sphinx (named 'sphinx.ext.*') or your custom ones.
extensions = [
    "sphinx.ext.napoleon",
    "sphinx.ext.autodoc",
    "sphinx.ext.doctest",
    "sphinx.ext.linkcode",
    "sphinx.ext.mathjax",
    "sphinx.ext.extlinks",
    "sphinx_math_dollar",
    "sphinx.ext.todo",
    "hoverxref.extension",  # This only works on read the docs
    "sphinx_design",
    "sphinxcontrib.bibtex",
    "nbsphinx",
    # see https://github.com/spatialaudio/nbsphinx/issues/24 for an explanation why this extension is necessary
    "IPython.sphinxext.ipython_console_highlighting",
    # Custom extensions
    "copy_notebooks",
]

# sphinx_math_dollar
mathjax3_config = {
    "tex": {
        "inlineMath": [["\\(", "\\)"]],
        "displayMath": [["\\[", "\\]"]],
    }
}

extlinks_detect_hardcoded_links = True
extlinks = {
    "gh": ("https://github.com/appliedAI-Initiative/pyDVL/%s", "GitHub %s"),
    "issue": ("https://github.com/appliedAI-Initiative/pyDVL/issues/%s", "issue %s"),
    "tfl": ("https://transferlab.appliedai.de/%s", "%s"),
}

bibtex_bibfiles = ["pydvl.bib"]
bibtex_bibliography_header = "References\n=========="
bibtex_footbibliography_header = bibtex_bibliography_header

# NBSphinx

# This is processed by Jinja2 and inserted before each notebook
nbsphinx_prolog = r"""
{% set docname = env.doc2path(env.docname, base=False).replace('examples', 'notebooks') %}

.. raw:: html

    <div class="admonition note">
        This page was generated from
        <a class="reference external" href="https://github.com/appliedAI-Initiative/pyDVL/blob/develop/{{ docname|e }}">{{ docname|e }}</a>
        <br>
        Interactive online version:
        <span style="white-space: nowrap;">
            <a href="https://mybinder.org/v2/gh/appliedAI-Initiative/pyDVL/develop?filepath={{ docname|e }}">
                <img alt="Binder badge" src="https://mybinder.org/badge_logo.svg" style="vertical-align:text-bottom">
            </a>
        </span>
    </div>

    <style>
        .nbinput .prompt,
        .nboutput .prompt {
            display: none;
        }
        @media not print {
            [data-theme='dark'] .output_area img {
                filter: invert(0.9);
            }
            @media (prefers-color-scheme: dark) {
                :root:not([data-theme="light"]) .output_area img {
                    filter: invert(0.9);
                }
            }
        }
    </style>
"""

# Display todos by setting to True
todo_include_todos = True


# adding links to source files (this works for gitlab and github like hosts and might need to be adjusted for others)
# see https://www.sphinx-doc.org/en/master/usage/extensions/linkcode.html#module-sphinx.ext.linkcode
def linkcode_resolve(domain, info):
    link_prefix = "https://github.com/appliedAI-Initiative/pyDVL/blob/develop"
    if domain != "py":
        return None
    if not info["module"]:
        return None

    path, link_extension = get_path_and_link_extension(info["module"])
    object_name = info["fullname"]
    if (
        "." in object_name
    ):  # don't add source link to methods within classes (you might want to change that)
        return None
    lineno = lineno_from_object_name(path, object_name)
    return f"{link_prefix}/{link_extension}#L{lineno}"


def get_path_and_link_extension(module: str):
    """
    :return: tuple of the form (path, link_extension) where
        the first entry is the local path to a given module or to __init__.py of the package
        and the second entry is the corresponding path from the top level directory
    """
    filename = module.replace(".", "/")
    docs_dir = os.path.dirname(os.path.realpath(__file__))
    source_path_prefix = os.path.join(docs_dir, f"../src/{filename}")

    if os.path.exists(source_path_prefix + ".py"):
        link_extension = f"src/{filename}.py"
        return source_path_prefix + ".py", link_extension
    elif os.path.exists(os.path.join(source_path_prefix, "__init__.py")):
        link_extension = f"src/{filename}/__init__.py"
        return os.path.join(source_path_prefix, "__init__.py"), link_extension
    else:
        raise Exception(
            f"{source_path_prefix} is neither a module nor a package with init - "
            f"did you forget to add an __init__.py?"
        )


def lineno_from_object_name(source_file, object_name):
    desired_node_name = object_name.split(".")[0]
    with open(source_file, "r") as f:
        source_node = ast.parse(f.read())
    desired_node = next(
        (
            node
            for node in source_node.body
            if getattr(node, "name", "") == desired_node_name
        ),
        None,
    )
    if desired_node is None:
        logger.warning(f"Could not find object {desired_node_name} in {source_file}")
        return 0
    else:
        return desired_node.lineno


# this is useful for keeping the docs build environment small. Add heavy requirements here
# and all other requirements to docs/requirements.txt
autodoc_mock_imports = []

autodoc_default_options = {
    "exclude-members": "log",
    "member-order": "bysource",
    "show-inheritance": True,
}

# Add any paths that contain templates here, relative to this directory.
templates_path = ["_templates"]

# The suffix of source filenames.
source_suffix = ".rst"

# The encoding of source files.
# source_encoding = 'utf-8-sig'

# The master toctree document.
master_doc = "index"

# General information about the project.
project = "pyDVL"

# The version info for the project you're documenting, acts as replacement for
# |version| and |release|, also used in various other places throughout the
# built documents.
#
# The full version, including alpha/beta/rc tags.
version = pkg_resources.get_distribution(project).version
release = version
# The short X.Y version.
major_v, minor_v = version.split(".")[:2]
version = f"{major_v}.{minor_v}"

# The language for content autogenerated by Sphinx. Refer to documentation
# for a list of supported languages.
# language = None

# There are two options for replacing |today|: either, you set today to some
# non-false value, then it is used:
# today = ''
# Else, today_fmt is used as the format for a strftime call.
# today_fmt = '%B %d, %Y'

# List of patterns, relative to source directory, that match files and
# directories to ignore when looking for source files.
exclude_patterns = ["_build"]

# The reST default role (used for this markup: `text`) to use for all documents.
# default_role = None

# If true, '()' will be appended to :func: etc. cross-reference text.
# add_function_parentheses = True

# If true, the current module name will be prepended to all description
# unit titles (such as .. function::).
add_module_names = False

# If true, sectionauthor and moduleauthor directives will be shown in the
# output. They are ignored by default.
# show_authors = False

# The name of the Pygments (syntax highlighting) style to use.
pygments_style = "sphinx"

# A list of ignored prefixes for module index sorting.
# modindex_common_prefix = []


# -- Options for HTML output ---------------------------------------------------

# Add a tooltip to all :ref: roles
# This requires a backend server to retrieve the tooltip content. As of Nov 22,
# sphinx-hoverxref only supports Read the Docs as backend.
# See https://sphinx-hoverxref.readthedocs.io/en/latest/configuration.html
# for further configuration options
# hoverxref_auto_ref = True

# The theme to use for HTML and HTML Help pages.  See the documentation for
# a list of builtin themes.
html_theme = "furo"

# Furo theme options:
html_theme_options = {
    "sidebar_hide_name": True,
    "navigation_with_keys": True,
    "announcement": "pyDVL is in an early stage of development. Expect changes to functionality and the API until version 1.0.0.",
    "footer_icons": [
        {
            "name": "GitHub",
            "url": "https://github.com/appliedAI-Initiative/pyDVL",
            "html": """
                <svg stroke="currentColor" fill="currentColor" stroke-width="0" viewBox="0 0 16 16">
                    <path fill-rule="evenodd" d="M8 0C3.58 0 0 3.58 0 8c0 3.54 2.29 6.53 5.47 7.59.4.07.55-.17.55-.38 0-.19-.01-.82-.01-1.49-2.01.37-2.53-.49-2.69-.94-.09-.23-.48-.94-.82-1.13-.28-.15-.68-.52-.01-.53.63-.01 1.08.58 1.23.82.72 1.21 1.87.87 2.33.66.07-.52.28-.87.51-1.07-1.78-.2-3.64-.89-3.64-3.95 0-.87.31-1.59.82-2.15-.08-.2-.36-1.02.08-2.12 0 0 .67-.21 2.2.82.64-.18 1.32-.27 2-.27.68 0 1.36.09 2 .27 1.53-1.04 2.2-.82 2.2-.82.44 1.1.16 1.92.08 2.12.51.56.82 1.27.82 2.15 0 3.07-1.87 3.75-3.65 3.95.29.25.54.73.54 1.48 0 1.07-.01 1.93-.01 2.2 0 .21.15.46.55.38A8.013 8.013 0 0 0 16 8c0-4.42-3.58-8-8-8z"></path>
                </svg>
            """,
            "class": "",
        },
    ],
}

# Add any paths that contain custom themes here, relative to this directory.
# html_theme_path = []

# The name for this set of Sphinx documents.  If None, it defaults to
# "<project> v<release> documentation".
# html_title = None

# A shorter title for the navigation bar.  Default is the same as html_title.
# html_short_title = None

# The name of an image file (relative to this directory) to place at the top
# of the sidebar.
html_logo = os.fspath(ROOT_DIR.joinpath("logo.svg"))

# The name of an image file (within the static path) to use as favicon of the
# docs.  This file should be a Windows icon file (.ico) being 16x16 or 32x32
# pixels large.
# html_favicon = None

# Add any paths that contain custom static files (such as style sheets) here,
# relative to this directory. They are copied after the builtin static files,
# so a file named "default.css" will overwrite the builtin "default.css".
html_static_path = []

# If not '', a 'Last updated on:' timestamp is inserted at every page bottom,
# using the given strftime format.
# html_last_updated_fmt = '%b %d, %Y'

# If true, SmartyPants will be used to convert quotes and dashes to
# typographically correct entities.
# html_use_smartypants = True

# Custom sidebar templates, maps document names to template names.
# html_sidebars = {}

# Additional templates that should be rendered to pages, maps page names to
# template names.
# html_additional_pages = {}

# If false, no module index is generated.
# html_domain_indices = True

# If false, no index is generated.
# html_use_index = True

# If true, the index is split into individual pages for each letter.
# html_split_index = False

# If true, links to the reST sources are added to the pages.
# html_show_sourcelink = True

# If true, "Created using Sphinx" is shown in the HTML footer. Default is True.
# html_show_sphinx = True

# If true, "(C) Copyright ..." is shown in the HTML footer. Default is True.
html_show_copyright = True
copyright = "AppliedAI Institute gGmbH"

# If true, an OpenSearch description file will be output, and all pages will
# contain a <link> tag referring to it.  The value of this option must be the
# base URL from which the finished HTML is served.
# html_use_opensearch = ''

# This is the file name suffix for HTML files (e.g. ".xhtml").
# html_file_suffix = None

# Output file base name for HTML help builder.
htmlhelp_basename = "pydvl_doc"


# -- Options for LaTeX output --------------------------------------------------

latex_elements = {
    # The paper size ('letterpaper' or 'a4paper').
    # 'papersize': 'letterpaper',
    # The font size ('10pt', '11pt' or '12pt').
    # 'pointsize': '10pt',
    # Additional stuff for the LaTeX preamble.
    # 'preamble': '',
}

# Grouping the document tree into LaTeX files. List of tuples
# (source start file, target name, title, author, documentclass [howto/manual]).
# latex_documents = []

# The name of an image file (relative to this directory) to place at the top of
# the title page.
# latex_logo = None

# For "manual" documents, if this is true, then toplevel headings are parts,
# not chapters.
# latex_use_parts = False

# If true, show page references after internal links.
# latex_show_pagerefs = False

# If true, show URL addresses after external links.
# latex_show_urls = False

# Documents to append as an appendix to all manuals.
# latex_appendices = []

# If false, no module index is generated.
# latex_domain_indices = True


# -- Options for manual page output --------------------------------------------

# One entry per manual page. List of tuples
# (source start file, name, description, authors, manual section).
man_pages = [
    (
        "index",
        "pydvl",
        "",
        ["appliedAI"],
        1,
    )
]

# If true, show URL addresses after external links.
# man_show_urls = False


# -- Options for Texinfo output ------------------------------------------------

# Grouping the document tree into Texinfo files. List of tuples
# (source start file, target name, title, author,
#  dir menu entry, description, category)
# texinfo_documents = []

# Documents to append as an appendix to all manuals.
# texinfo_appendices = []

# If false, no module index is generated.
# texinfo_domain_indices = True

# How to display URL addresses: 'footnote', 'no', or 'inline'.
# texinfo_show_urls = 'footnote'



from __future__ import annotations

import functools
import operator
import pickle
from copy import deepcopy

import cloudpickle
import numpy as np
import pytest

from pydvl.utils.status import Status
from pydvl.value import ValuationResult


@pytest.fixture
def dummy_values(values, names):
    return ValuationResult(
        algorithm="dummy_valuator",
        status=Status.Converged,
        values=np.array(values),
        variances=np.zeros_like(values),
        data_names=names,
        sort=True,
    )


@pytest.mark.parametrize(
    "values, names, ranks_asc", [([], [], []), ([2, 3, 1], ["a", "b", "c"], [2, 0, 1])]
)
def test_sorting(values, names, ranks_asc, dummy_values):
    dummy_values.sort(key="value")
    assert np.alltrue([it.value for it in dummy_values] == sorted(values))
    assert np.alltrue(dummy_values.indices == ranks_asc)
    assert np.alltrue(
        [it.value for it in reversed(dummy_values)] == sorted(values, reverse=True)
    )

    dummy_values.sort(reverse=True)
    assert np.alltrue([it.value for it in dummy_values] == sorted(values, reverse=True))
    assert np.alltrue(dummy_values.indices == list(reversed(ranks_asc)))

    dummy_values.sort(key="index")
    assert np.alltrue(dummy_values.indices == list(range(len(values))))
    assert np.alltrue([it.value for it in dummy_values] == values)


@pytest.mark.parametrize(
    "values, names, ranks_asc", [([], [], []), ([2, 3, 1], ["a", "b", "c"], [2, 0, 1])]
)
def test_dataframe_sorting(values, names, ranks_asc, dummy_values):
    sorted_names = [names[r] for r in ranks_asc]
    try:
        import pandas

        df = dummy_values.to_dataframe(use_names=False)
        assert np.alltrue(df.index.values == ranks_asc)

        df = dummy_values.to_dataframe(use_names=True)
        assert np.alltrue(df.index.values == sorted_names)
        assert np.alltrue(df["dummy_valuator"].values == sorted(values))

        dummy_values.sort(reverse=True)
        df = dummy_values.to_dataframe(use_names=True)
        assert np.alltrue(df.index.values == list(reversed(sorted_names)))
        assert np.alltrue(df["dummy_valuator"].values == sorted(values, reverse=True))
    except ImportError:
        pass


@pytest.mark.parametrize(
    "values, names, ranks_asc", [([], [], []), ([2, 3, 1], ["a", "b", "c"], [2, 0, 1])]
)
def test_iter(names, ranks_asc, dummy_values):
    for rank, it in enumerate(dummy_values):
        assert it.index == ranks_asc[rank]

    for rank, it in enumerate(dummy_values):
        assert it.name == names[ranks_asc[rank]]


@pytest.mark.parametrize(
    "values, names, ranks_asc", [([], [], []), ([2, 3, 1], ["a", "b", "c"], [2, 0, 1])]
)
def test_todataframe(ranks_asc, dummy_values):
    df = dummy_values.to_dataframe()
    assert "dummy_valuator" in df.columns
    assert "dummy_valuator_stderr" in df.columns
    assert np.alltrue(df.index.values == ranks_asc)

    df = dummy_values.to_dataframe(column="val")
    assert "val" in df.columns
    assert "val_stderr" in df.columns
    assert np.alltrue(df.index.values == ranks_asc)

    df = dummy_values.to_dataframe(use_names=True)
    assert np.alltrue(df.index.values == [it.name for it in dummy_values])


@pytest.mark.parametrize(
    "values, names, ranks_asc",
    [([], [], []), ([2.0, 3.0, 1.0, 6.0], ["a", "b", "c", "d"], [2, 0, 1, 3])],
)
def test_indexing(ranks_asc, dummy_values):
    if len(ranks_asc) == 0:
        with pytest.raises(IndexError):
            dummy_values[1]  # noqa
        dummy_values[:2]  # noqa
    else:
        assert ranks_asc[:] == [it.index for it in dummy_values[:]]
        assert ranks_asc[0] == dummy_values[0].index
        assert [ranks_asc[0]] == [it.index for it in dummy_values[[0]]]
        assert ranks_asc[:2] == [it.index for it in dummy_values[:2]]
        assert ranks_asc[:2] == [it.index for it in dummy_values[[0, 1]]]
        assert ranks_asc[:-2] == [it.index for it in dummy_values[:-2]]
        assert ranks_asc[-2:] == [it.index for it in dummy_values[-2:]]
        assert ranks_asc[-2:] == [it.index for it in dummy_values[[-2, -1]]]


def test_get_idx():
    """Test getting by data index"""
    values = np.array([5.0, 2.0, 3.0])
    indices = np.array([3, 4, 2])
    result = ValuationResult(values=values, indices=indices)
    with pytest.raises(IndexError):
        result.get(5)
    for v, idx in zip(values, indices):
        assert v == result.get(idx).value
    result.sort()
    for v, idx in zip(values, indices):
        assert v == result.get(idx).value


def test_updating():
    # Test simple updating
    v = ValuationResult(values=np.array([1.0, 2.0]))
    v.update(0, 1.0)
    assert v.values[0] == 1.0
    assert v.counts[0] == 2

    v.update(1, 4.0)
    assert v.values[1] == 3.0
    assert v.variances[1] == 1.0

    v.update(1, 3.0)
    assert v.values[1] == 3.0
    assert np.isclose(v.variances[1], 2 / 3)

    # Test after sorting
    v = ValuationResult(values=np.array([3.0, 1.0]))
    v.sort()
    v.update(0, 1.0)
    assert v.values[0] == 2.0

    # Test data indexing
    v = ValuationResult(values=np.array([3.0, 1.0]), indices=np.array([3, 4]))
    v.update(4, 1.0)
    assert v.values[1] == 1.0
    assert v.counts[1] == 2


@pytest.mark.parametrize(
    "serialize, deserialize",
    [(pickle.dumps, pickle.loads), (cloudpickle.dumps, cloudpickle.loads)],
)
@pytest.mark.parametrize("values, names", [([], None), ([2.0, 3.0, 1.0], None)])
def test_serialization(serialize, deserialize, dummy_values):
    serded = deserialize(serialize(dummy_values))
    assert dummy_values == serded  # Serialization OK (if __eq__ ok...)
    dummy_values.sort(reverse=True)
    assert dummy_values != serded  # Order checks


@pytest.mark.parametrize("values, names", [([], []), ([2, 3, 1], ["a", "b", "c"])])
def test_equality(values, names, dummy_values):
    assert dummy_values == dummy_values

    c = deepcopy(dummy_values)
    dummy_values.sort(reverse=True)
    assert c != dummy_values

    c2 = ValuationResult(
        algorithm="dummy",
        status=c.status,
        values=c.values,
        variances=c._variances,
        data_names=c._names,
    )
    assert c != c2

    c2 = ValuationResult(
        algorithm=c._algorithm,
        status=Status.Failed,
        values=c.values,
        variances=c.variances,
        data_names=c.names,
    )
    assert c != c2

    c2 = ValuationResult(
        algorithm=c._algorithm,
        status=c._status,
        values=c._values,
        variances=c._variances,
        data_names=c._names,
    )
    c2.sort(c._sort_order)

    assert c == c2

    if len(c) > 0:
        c2 = ValuationResult(
            algorithm=c._algorithm,
            status=c._status,
            values=c._values + 1.0,
            variances=c._variances,
            data_names=c._names,
        )
        assert c != c2


@pytest.mark.parametrize(
    "extra_values", [{"test_value": 1.2}, {"test_value1": 1.2, "test_value2": "test"}]
)
def test_extra_values(extra_values):
    kwargs = dict(
        algorithm="test",
        status=Status.Converged,
        values=np.random.rand(10),
        sort=True,
        test_value=1.2,
    )
    kwargs.update(extra_values)
    result = ValuationResult(**kwargs)
    for k, v in extra_values.items():
        assert getattr(result, k) == v
    # Making sure that the repr dunder method works when using extra values
    repr_string = repr(result)
    for k, v in extra_values.items():
        assert k in repr_string


@pytest.mark.parametrize("size", [1, 10])
@pytest.mark.parametrize("total", [None, 1.0, -1.0])
def test_from_random_creation(size: int, total: float | None):
    result = ValuationResult.from_random(size=size, total=total)
    assert len(result) == size
    assert result.status == Status.Converged
    assert result.algorithm == "random"
    if total is not None:
        assert np.isclose(np.sum(result.values), total)


def test_from_random_creation_errors():
    with pytest.raises(ValueError):
        ValuationResult.from_random(size=0)


def test_adding_random():
    """Test adding multiple valuation results together.

    First we generate a matrix of values, then we split it into multiple subsets
    and create a valuation result for each subset. Then we add all the valuation
    results together and check that the resulting means and variances match with
    those of the original matrix.
    """
    n_samples, n_values, n_subsets = 10, 1000, 12
    values = np.random.rand(n_samples, n_values)
    split_indices = np.sort(np.random.randint(1, n_values, size=n_subsets - 1))
    splits = np.split(values, split_indices, axis=1)
    vv = [
        ValuationResult(
            algorithm="dummy",
            status=Status.Pending,
            values=np.average(s, axis=1),
            variances=np.var(s, axis=1),
            counts=s.shape[1] * np.ones(n_samples),
        )
        for s in splits
    ]
    result: ValuationResult = functools.reduce(operator.add, vv)

    true_means = values.mean(axis=1)
    true_variances = values.var(axis=1)

    assert np.allclose(true_means[result.indices], result.values)
    assert np.allclose(true_variances[result.indices], result.variances)


@pytest.mark.parametrize(
    "indices_1, names_1, values_1, indices_2, names_2, values_2,"
    "expected_indices, expected_names, expected_values",
    [  # Disjoint indices
        (
            [0, 1, 2],
            ["a", "b", "c"],
            [0, 1, 2],
            [3, 4, 5],
            ["d", "e", "f"],
            [3, 4, 5],
            [0, 1, 2, 3, 4, 5],
            ["a", "b", "c", "d", "e", "f"],
            [0, 1, 2, 3, 4, 5],
        ),
        # Overlapping indices (recall that values are running averages)
        (
            [0, 1, 2],
            ["a", "b", "c"],
            [0, 1, 2],
            [1, 2, 3],
            ["b", "c", "d"],
            [3, 4, 5],
            [0, 1, 2, 3],
            ["a", "b", "c", "d"],
            [0, 2, 3, 5],
        ),
        # Overlapping indices with different lengths
        (
            [0, 1, 2],
            ["a", "b", "c"],
            [0, 1, 2],
            [1, 2],
            ["b", "c"],
            [3, 4],
            [0, 1, 2],
            ["a", "b", "c"],
            [0, 2, 3],
        ),
    ],
)
def test_adding_different_indices(
    indices_1,
    names_1,
    values_1,
    indices_2,
    names_2,
    values_2,
    expected_indices,
    expected_names,
    expected_values,
):
    """Test adding valuation results with disjoint indices"""
    v1 = ValuationResult(
        indices=np.array(indices_1), values=np.array(values_1), data_names=names_1
    )
    v2 = ValuationResult(
        indices=np.array(indices_2), values=np.array(values_2), data_names=names_2
    )
    v3 = v1 + v2

    assert np.allclose(v3.indices, np.array(expected_indices))
    assert np.allclose(v3.values, np.array(expected_values))
    assert np.all(v3.names == expected_names)


@pytest.mark.parametrize(
    "indices, index_t, data_names, name_t",
    [
        ([0, 1, 2], np.int32, ["a", "b", "c"], "<U1"),
        ([4, 1, 7], np.int64, [4, 1, 7], np.int64),
        ([4, 1, 7], np.int32, [4, 1, 7], np.float_),
    ],
)
def test_types(indices, index_t, data_names, name_t):
    """Test that types for indices and names are correctly preserved when adding
    valuation results"""

    v = ValuationResult(
        indices=np.array(indices, dtype=index_t),
        values=np.ones(len(indices), dtype=np.float_),
        data_names=np.array(data_names, dtype=name_t),
    )
    assert v.indices.dtype == index_t
    assert v.names.dtype == name_t

    v2 = ValuationResult(
        indices=np.array(indices),
        values=np.ones(len(indices)),
        variances=np.zeros(len(indices)),
        data_names=data_names,
    )
    v += v2
    assert v.indices.dtype == index_t
    assert v.names.dtype == name_t


@pytest.mark.parametrize("data_names", [["a", "b", "c"]])
def test_names(data_names):
    """Test that data names are preserved after addition of results"""

    n = len(data_names)
    v = ValuationResult.from_random(size=n, data_names=data_names)
    v2 = ValuationResult.from_random(size=n, data_names=data_names)

    v += v2
    assert np.all(v.names == np.array(data_names))


@pytest.mark.parametrize("n", [0, 5])
def test_empty(n):
    v = ValuationResult.empty()
    assert len(v) == 0
    v2 = ValuationResult(values=np.arange(n))
    v += v2
    assert len(v2) == n


def test_empty_deprecation():
    with pytest.warns(DeprecationWarning):
        v3 = ValuationResult.empty(indices=[1, 2, 3])



import itertools
from typing import List

import numpy as np
import pytest

from pydvl.influence.conjugate_gradient import (
    batched_preconditioned_conjugate_gradient,
    conjugate_gradient_condition_number_based_error_bound,
)


class AlgorithmTestSettings:
    A_NORM_TOL: float = 1e-4
    ACCEPTABLE_FAILED_PERC_A_NORM: float = 1e-2
    ACCEPTABLE_FAILED_PERC_BOUND: float = 1e-2
    BOUND_TOL: float = 1e-4
    CG_DAMPING: float = 1e-10

    CG_TEST_CONDITION_NUMBERS: List[int] = [10]
    CG_TEST_BATCH_SIZES: List[int] = [16, 32]
    CG_TEST_DIMENSIONS: List[int] = [2, 20, 60, 100]


test_cases = list(
    itertools.product(
        AlgorithmTestSettings.CG_TEST_DIMENSIONS,
        AlgorithmTestSettings.CG_TEST_BATCH_SIZES,
        AlgorithmTestSettings.CG_TEST_CONDITION_NUMBERS,
    )
)


def lmb_test_case_to_str(packed_i_test_case):
    i, test_case = packed_i_test_case
    return f"Problem #{i} of dimension {test_case[0]} with batch size {test_case[1]} and condition number"


test_case_ids = list(map(lmb_test_case_to_str, zip(range(len(test_cases)), test_cases)))


@pytest.mark.parametrize(
    "problem_dimension,batch_size,condition_number",
    test_cases,
    ids=test_case_ids,
    indirect=True,
)
def test_conjugate_gradients_mvp(quadratic_linear_equation_system):
    A, b = quadratic_linear_equation_system
    x0 = np.zeros_like(b)
    xn, n = batched_preconditioned_conjugate_gradient(A, b, x0=x0, rtol=10e-7)
    check_solution(A, b, n, x0, xn)


@pytest.mark.parametrize(
    "problem_dimension,batch_size,condition_number",
    test_cases,
    ids=test_case_ids,
    indirect=True,
)
def test_conjugate_gradients_fn(quadratic_linear_equation_system):
    A, b = quadratic_linear_equation_system
    new_A = np.copy(A)
    A = lambda v: v @ new_A.T
    x0 = np.zeros_like(b)
    xn, n = batched_preconditioned_conjugate_gradient(A, b, x0=x0, rtol=10e-7)
    check_solution(new_A, b, n, x0, xn)


@pytest.mark.parametrize(
    "problem_dimension,batch_size,condition_number",
    test_cases,
    ids=test_case_ids,
    indirect=True,
)
def test_conjugate_gradients_mvp_preconditioned(quadratic_linear_equation_system):
    A, b = quadratic_linear_equation_system
    x0 = np.zeros_like(b)
    xn, n = batched_preconditioned_conjugate_gradient(A, b, x0=x0, rtol=10e-7)
    check_solution(A, b, n, x0, xn)


def check_solution(A, b, n, x0, xn):
    """
    Uses standard inversion techniques to verify the solution of the problem. It checks:

    - That the solution is not nan at all positions.
    - The solution fulfills an error bound, which depends on A and the number of iterations.
    - Only a certain percentage of the batch is allowed to be false.
    """
    assert np.all(np.logical_not(np.isnan(xn)))
    inv_A = np.linalg.pinv(A)

    xt = b @ inv_A.T
    bound = conjugate_gradient_condition_number_based_error_bound(A, n, x0, xt)
    norm_A = lambda v: np.sqrt(np.einsum("ia,ab,ib->i", v, A, v))
    assert np.all(np.logical_not(np.isnan(xn)))

    error = norm_A(xt - xn)
    failed = error > bound + AlgorithmTestSettings.BOUND_TOL
    realized_failed_percentage = np.sum(failed) / len(failed)
    assert (
        realized_failed_percentage < AlgorithmTestSettings.ACCEPTABLE_FAILED_PERC_BOUND
    )

    failed = error > AlgorithmTestSettings.A_NORM_TOL
    realized_failed_percentage = np.sum(failed) / len(failed)
    assert (
        realized_failed_percentage < AlgorithmTestSettings.ACCEPTABLE_FAILED_PERC_A_NORM
    )



import os
import shutil
from pathlib import Path

from sphinx.application import Sphinx
from sphinx.config import Config
from sphinx.util import logging

logger = logging.getLogger(__name__)


def copy_notebooks(app: Sphinx, config: Config) -> None:
    logger.info("Copying notebooks to examples directory")
    root_dir = Path(app.confdir).parent
    notebooks_dir = root_dir / "notebooks"
    docs_examples_dir = root_dir / "docs" / "examples"
    notebook_filepaths = list(notebooks_dir.glob("*.ipynb"))
    for notebook in notebook_filepaths:
        target_filepath = docs_examples_dir / notebook.name
        try:
            if os.path.getmtime(notebook) <= os.path.getmtime(target_filepath):
                logger.info(
                    f"Notebook '{os.fspath(notebook)}' hasn't been updated, skipping."
                )
                continue
        except FileNotFoundError:
            pass
        logger.info(
            f"Copying '{os.fspath(notebook)}' to '{os.fspath(target_filepath)}'"
        )
        shutil.copyfile(src=notebook, dst=target_filepath)
    logger.info("Finished copying notebooks to examples directory")


def setup(app):
    app.connect("config-inited", copy_notebooks)



""" This module contains types, protocols, decorators and generic function
transformations. Some of it probably belongs elsewhere.
"""
import inspect
from typing import Callable, Protocol, Type

from numpy.typing import NDArray

__all__ = ["SupervisedModel"]


class SupervisedModel(Protocol):
    """This is the minimal Protocol that valuation methods require from
    models in order to work.

    All that is needed are the standard sklearn methods `fit()`, `predict()` and
    `score()`.
    """

    def fit(self, x: NDArray, y: NDArray):
        pass

    def predict(self, x: NDArray) -> NDArray:
        pass

    def score(self, x: NDArray, y: NDArray) -> float:
        pass


def maybe_add_argument(fun: Callable, new_arg: str):
    """Wraps a function to accept the given keyword parameter if it doesn't
    already.

    If `fun` already takes a keyword parameter of name `new_arg`, then it is
    returned as is. Otherwise, a wrapper is returned which merely ignores the
    argument.

    :param fun: The function to wrap
    :param new_arg: The name of the argument that the new function will accept
        (and ignore).
    :return: A new function accepting one more keyword argument.
    """
    params = inspect.signature(fun).parameters
    if new_arg in params.keys():
        return fun

    def wrapper(*args, **kwargs):
        try:
            del kwargs[new_arg]
        except KeyError:
            pass
        return fun(*args, **kwargs)

    return wrapper



from concurrent.futures import Executor, ThreadPoolExecutor
from contextlib import contextmanager
from typing import Generator, Optional

from pydvl.utils.config import ParallelConfig
from pydvl.utils.parallel.futures.ray import RayExecutor

__all__ = ["init_executor"]


@contextmanager
def init_executor(
    max_workers: Optional[int] = None,
    config: ParallelConfig = ParallelConfig(),
    **kwargs,
) -> Generator[Executor, None, None]:
    """Initializes a futures executor based on the passed parallel configuration object.

    :param max_workers: Maximum number of concurrent tasks.
    :param config: instance of :class:`~pydvl.utils.config.ParallelConfig` with cluster address, number of cpus, etc.
    :param kwargs: Other optional parameter that will be passed to the executor.

    :Example:

    >>> from pydvl.utils.parallel.futures import init_executor
    >>> from pydvl.utils.config import ParallelConfig
    >>> config = ParallelConfig(backend="ray")
    >>> with init_executor(max_workers=3, config=config) as executor:
    ...     pass

    >>> from pydvl.utils.parallel.futures import init_executor
    >>> with init_executor() as executor:
    ...     future = executor.submit(lambda x: x + 1, 1)
    ...     result = future.result()
    ...
    >>> print(result)
    2

    >>> from pydvl.utils.parallel.futures import init_executor
    >>> with init_executor() as executor:
    ...     results = list(executor.map(lambda x: x + 1, range(5)))
    ...
    >>> print(results)
    [1, 2, 3, 4, 5]

    """
    if config.backend == "ray":
        with RayExecutor(max_workers, config=config, **kwargs) as executor:
            yield executor
    elif config.backend == "sequential":
        with ThreadPoolExecutor(1) as executor:
            yield executor
    else:
        raise NotImplementedError(f"Unexpected parallel type {config.backend}")



"""
Stopping criteria for value computations.

This module provides a basic set of stopping criteria, like :class:`MaxUpdates`,
:class:`MaxTime`, or :class:`HistoryDeviation` among others. These can behave in
different ways depending on the context. For example, :class:`MaxUpdates` limits
the number of updates to values, which depending on the algorithm may mean a
different number of utility evaluations or imply other computations like solving
a linear or quadratic program.

.. rubric:: Creating stopping criteria

The easiest way is to declare a function implementing the interface
:data:`StoppingCriterionCallable` and wrap it with :func:`make_criterion`. This
creates a :class:`StoppingCriterion` object that can be composed with other
stopping criteria.

Alternatively, and in particular if reporting of completion is required, one can
inherit from this class and implement the abstract methods
:meth:`~pydvl.value.stopping.StoppingCriterion._check` and
:meth:`~pydvl.value.stopping.StoppingCriterion.completion`.

.. rubric:: Composing stopping criteria

Objects of type :class:`StoppingCriterion` can be composed with the binary
operators ``&`` (*and*), and ``|`` (*or*), following the truth tables of
:class:`~pydvl.utils.status.Status`. The unary operator ``~`` (*not*) is also
supported. See :class:`StoppingCriterion` for details on how these operations
affect the behavior of the stopping criteria.
"""

import abc
import logging
from time import time
from typing import Callable, Optional, Type

import numpy as np
from deprecate import deprecated, void
from numpy.typing import NDArray

from pydvl.utils import Status
from pydvl.value import ValuationResult

__all__ = [
    "make_criterion",
    "AbsoluteStandardError",
    "StoppingCriterion",
    "StandardError",
    "MaxChecks",
    "MaxUpdates",
    "MinUpdates",
    "MaxTime",
    "HistoryDeviation",
]

logger = logging.getLogger(__name__)

StoppingCriterionCallable = Callable[[ValuationResult], Status]


class StoppingCriterion(abc.ABC):
    """A composable callable object to determine whether a computation
    must stop.

    A ``StoppingCriterion`` is a callable taking a
    :class:`~pydvl.value.result.ValuationResult` and returning a
    :class:`~pydvl.value.result.Status`. It also keeps track of individual
    convergence of values with :meth:`converged`, and reports the overall
    completion of the computation with :meth:`completion`.

    Instances of ``StoppingCriterion`` can be composed with the binary operators
    ``&`` (*and*), and ``|`` (*or*), following the truth tables of
    :class:`~pydvl.utils.status.Status`. The unary operator ``~`` (*not*) is
    also supported. These boolean operations act according to the following
    rules:

    - The results of :meth:`_check` are combined with the operator. See
      :class:`~pydvl.utils.status.Status` for the truth tables.
    - The results of :meth:`converged` are combined with the operator (returning
      another boolean array).
    - The :meth:`completion` method returns the min, max, or the complement to 1
      of the completions of the operands, for AND, OR and NOT respectively. This
      is required for cases where one of the criteria does not keep track of the
      convergence of single values, e.g. :class:`MaxUpdates`, because
      :meth:`completion` by default returns the mean of the boolean convergence
      array.

    .. rubric:: Subclassing

    Subclassing this class requires implementing a :meth:`_check` method that
    returns a :class:`~pydvl.utils.status.Status` object based on a given
    :class:`~pydvl.value.result.ValuationResult`. This method should update the
    :attr:`converged` attribute, which is a boolean array indicating whether
    the value for each index has converged. When this is not possible,
    :meth:`completion` should be overridden to provide an overall completion
    value, since the default implementation returns the mean of :attr:`converged`.

    :param modify_result: If ``True`` the status of the input
        :class:`~pydvl.value.result.ValuationResult` is modified in place after
        the call.
    """

    # A boolean array indicating whether the corresponding element has converged
    _converged: NDArray[np.bool_]

    def __init__(self, modify_result: bool = True):
        self.modify_result = modify_result
        self._converged = np.full(0, False)

    @abc.abstractmethod
    def _check(self, result: ValuationResult) -> Status:
        """Check whether the computation should stop."""
        ...

    def completion(self) -> float:
        """Returns a value between 0 and 1 indicating the completion of the
        computation.
        """
        if self.converged.size == 0:
            return 0.0
        return np.mean(self.converged).item()

    @property
    def converged(self) -> NDArray[np.bool_]:
        """Returns a boolean array indicating whether the values have converged
        for each data point.

        Inheriting classes must set the ``_converged`` attribute in their
        :meth:`_check`.
        """
        return self._converged

    @property
    def name(self):
        return type(self).__name__

    def __call__(self, result: ValuationResult) -> Status:
        """Calls :meth:`_check`, maybe updating the result."""
        if len(result) == 0:
            logger.warning(
                "At least one iteration finished but no results where generated. "
                "Please check that your scorer and utility return valid numbers."
            )
        status = self._check(result)
        if self.modify_result:  # FIXME: this is not nice
            result._status = status
        return status

    def __and__(self, other: "StoppingCriterion") -> "StoppingCriterion":
        return make_criterion(
            fun=lambda result: self._check(result) & other._check(result),
            converged=lambda: self.converged & other.converged,
            completion=lambda: min(self.completion(), other.completion()),
            name=f"Composite StoppingCriterion: {self.name} AND {other.name}",
        )(modify_result=self.modify_result or other.modify_result)

    def __or__(self, other: "StoppingCriterion") -> "StoppingCriterion":
        return make_criterion(
            fun=lambda result: self._check(result) | other._check(result),
            converged=lambda: self.converged | other.converged,
            completion=lambda: max(self.completion(), other.completion()),
            name=f"Composite StoppingCriterion: {self.name} OR {other.name}",
        )(modify_result=self.modify_result or other.modify_result)

    def __invert__(self) -> "StoppingCriterion":
        return make_criterion(
            fun=lambda result: ~self._check(result),
            converged=lambda: ~self.converged,
            completion=lambda: 1 - self.completion(),
            name=f"Composite StoppingCriterion: NOT {self.name}",
        )(modify_result=self.modify_result)


def make_criterion(
    fun: StoppingCriterionCallable,
    converged: Callable[[], NDArray[np.bool_]] = None,
    completion: Callable[[], float] = None,
    name: str = None,
) -> Type[StoppingCriterion]:
    """Create a new :class:`StoppingCriterion` from a function.
    Use this to enable simpler functions to be composed with bitwise operators

    :param fun: The callable to wrap.
    :param converged: A callable that returns a boolean array indicating what
        values have converged.
    :param completion: A callable that returns a value between 0 and 1 indicating
        the rate of completion of the computation. If not provided, the fraction
        of converged values is used.
    :param name: The name of the new criterion. If ``None``, the ``__name__`` of
        the function is used.
    :return: A new subclass of :class:`StoppingCriterion`.
    """

    class WrappedCriterion(StoppingCriterion):
        def __init__(self, modify_result: bool = True):
            super().__init__(modify_result=modify_result)
            self._name = name or fun.__name__

        def _check(self, result: ValuationResult) -> Status:
            return fun(result)

        @property
        def converged(self) -> NDArray[np.bool_]:
            if converged is None:
                return super().converged
            return converged()

        @property
        def name(self):
            return self._name

        def completion(self) -> float:
            if completion is None:
                return super().completion()
            return completion()

    return WrappedCriterion


class AbsoluteStandardError(StoppingCriterion):
    r"""Determine convergence based on the standard error of the values.

    If $s_i$ is the standard error for datum $i$ and $v_i$ its value, then this
    criterion returns :attr:`~pydvl.utils.status.Status.Converged` if
    $s_i < \epsilon$ for all $i$ and a threshold value $\epsilon \gt 0$.

    :param threshold: A value is considered to have converged if the standard
        error is below this value. A way of choosing it is to pick some
        percentage of the range of the values. For Shapley values this is the
        difference between the maximum and minimum of the utility function (to
        see this substitute the maximum and minimum values of the utility into
        the marginal contribution formula).
    :param fraction: The fraction of values that must have converged for the
        criterion to return :attr:`~pydvl.utils.status.Status.Converged`.
    :param burn_in: The number of iterations to ignore before checking for
        convergence. This is required because computations typically start with
        zero variance, as a result of using
        :meth:`~pydvl.value.result.ValuationResult.empty`. The default is set to
        an arbitrary minimum which is usually enough but may need to be
        increased.
    """

    def __init__(
        self,
        threshold: float,
        fraction: float = 1.0,
        burn_in: int = 4,
        modify_result: bool = True,
    ):
        super().__init__(modify_result=modify_result)
        self.threshold = threshold
        self.fraction = fraction
        self.burn_in = burn_in

    def _check(self, result: ValuationResult) -> Status:
        self._converged = (result.stderr < self.threshold) & (
            result.counts > self.burn_in
        )
        if np.mean(self._converged) >= self.fraction:
            return Status.Converged
        return Status.Pending


class StandardError(AbsoluteStandardError):
    @deprecated(target=AbsoluteStandardError, deprecated_in="0.6.0", remove_in="0.8.0")
    def __init__(self, *args, **kwargs):
        void(*args, **kwargs)


class MaxChecks(StoppingCriterion):
    """Terminate as soon as the number of checks exceeds the threshold.

    A "check" is one call to the criterion.

    :param n_checks: Threshold: if ``None``, no _check is performed,
        effectively creating a (never) stopping criterion that always returns
        ``Pending``.
    """

    def __init__(self, n_checks: Optional[int], modify_result: bool = True):
        super().__init__(modify_result=modify_result)
        if n_checks is not None and n_checks < 1:
            raise ValueError("n_iterations must be at least 1 or None")
        self.n_checks = n_checks
        self._count = 0

    def _check(self, result: ValuationResult) -> Status:
        if self.n_checks:
            self._count += 1
            if self._count > self.n_checks:
                self._converged = np.ones_like(result.values, dtype=bool)
                return Status.Converged
        return Status.Pending

    def completion(self) -> float:
        if self.n_checks:
            return min(1.0, self._count / self.n_checks)
        return 0.0


class MaxUpdates(StoppingCriterion):
    """Terminate if any number of value updates exceeds or equals the given
    threshold.

    This checks the ``counts`` field of a
    :class:`~pydvl.value.result.ValuationResult`, i.e. the number of times that
    each index has been updated. For powerset samplers, the maximum of this
    number coincides with the maximum number of subsets sampled. For permutation
    samplers, it coincides with the number of permutations sampled.

    :param n_updates: Threshold: if ``None``, no _check is performed,
        effectively creating a (never) stopping criterion that always returns
        ``Pending``.
    """

    def __init__(self, n_updates: Optional[int], modify_result: bool = True):
        super().__init__(modify_result=modify_result)
        if n_updates is not None and n_updates < 1:
            raise ValueError("n_updates must be at least 1 or None")
        self.n_updates = n_updates
        self.last_max = 0

    def _check(self, result: ValuationResult) -> Status:
        if self.n_updates:
            self._converged = result.counts >= self.n_updates
            try:
                self.last_max = int(np.max(result.counts))
                if self.last_max >= self.n_updates:
                    return Status.Converged
            except ValueError:  # empty counts array. This should not happen
                pass
        return Status.Pending

    def completion(self) -> float:
        if self.n_updates:
            return self.last_max / self.n_updates
        return 0.0


class MinUpdates(StoppingCriterion):
    """Terminate as soon as all value updates exceed or equal the given threshold.

    This checks the ``counts`` field of a
    :class:`~pydvl.value.result.ValuationResult`, i.e. the number of times that
    each index has been updated. For powerset samplers, the minimum of this
    number is a lower bound for the number of subsets sampled. For
    permutation samplers, it lower-bounds the amount of permutations sampled.

    :param n_updates: Threshold: if ``None``, no _check is performed,
        effectively creating a (never) stopping criterion that always returns
        ``Pending``.
    """

    def __init__(self, n_updates: Optional[int], modify_result: bool = True):
        super().__init__(modify_result=modify_result)
        self.n_updates = n_updates
        self.last_min = 0

    def _check(self, result: ValuationResult) -> Status:
        if self.n_updates is not None:
            self._converged = result.counts >= self.n_updates
            try:
                self.last_min = int(np.min(result.counts))
                if self.last_min >= self.n_updates:
                    return Status.Converged
            except ValueError:  # empty counts array. This should not happen
                pass
        return Status.Pending

    def completion(self) -> float:
        if self.n_updates:
            return self.last_min / self.n_updates
        return 0.0


class MaxTime(StoppingCriterion):
    """Terminate if the computation time exceeds the given number of seconds.

    Checks the elapsed time since construction

    :param seconds: Threshold: The computation is terminated if the elapsed time
        between object construction and a _check exceeds this value. If ``None``,
        no _check is performed, effectively creating a (never) stopping criterion
        that always returns ``Pending``.
    """

    def __init__(self, seconds: Optional[float], modify_result: bool = True):
        super().__init__(modify_result=modify_result)
        self.max_seconds = seconds or np.inf
        if self.max_seconds <= 0:
            raise ValueError("Number of seconds for MaxTime must be positive or None")
        self.start = time()

    def _check(self, result: ValuationResult) -> Status:
        if self._converged is None:
            self._converged = np.full(result.values.shape, False)
        if time() > self.start + self.max_seconds:
            self._converged.fill(True)
            return Status.Converged
        return Status.Pending

    def completion(self) -> float:
        if self.max_seconds is None:
            return 0.0
        return (time() - self.start) / self.max_seconds


class HistoryDeviation(StoppingCriterion):
    r"""A simple check for relative distance to a previous step in the
    computation.

    The method used by :footcite:t:`ghorbani_data_2019` computes the relative
    distances between the current values $v_i^t$ and the values at the previous
    checkpoint $v_i^{t-\tau}$. If the sum is below a given threshold, the
    computation is terminated.

    $$\sum_{i=1}^n \frac{\left| v_i^t - v_i^{t-\tau} \right|}{v_i^t} <
    \epsilon.$$

    When the denominator is zero, the summand is set to the value of $v_i^{
    t-\tau}$.

    This implementation is slightly generalised to allow for different number of
    updates to individual indices, as happens with powerset samplers instead of
    permutations. Every subset of indices that is found to converge can be
    pinned to that state. Once all indices have converged the method has
    converged.

    .. warning::
       This criterion is meant for the reproduction of the results in the paper,
       but we do not recommend using it in practice.

    :param n_steps: Checkpoint values every so many updates and use these saved
        values to compare.
    :param rtol: Relative tolerance for convergence ($\epsilon$ in the formula).
    :param pin_converged: If ``True``, once an index has converged, it is pinned
    """

    _memory: NDArray[np.float_]

    def __init__(
        self,
        n_steps: int,
        rtol: float,
        pin_converged: bool = True,
        modify_result: bool = True,
    ):
        super().__init__(modify_result=modify_result)
        if n_steps < 1:
            raise ValueError("n_steps must be at least 1")
        if rtol <= 0 or rtol >= 1:
            raise ValueError("rtol must be in (0, 1)")

        self.n_steps = n_steps
        self.rtol = rtol
        self.update_op = np.logical_or if pin_converged else np.logical_and
        self._memory = None  # type: ignore

    def _check(self, r: ValuationResult) -> Status:
        if self._memory is None:
            self._memory = np.full((len(r.values), self.n_steps + 1), np.inf)
            self._converged = np.full(len(r), False)
            return Status.Pending

        # shift left: last column is the last set of values
        self._memory = np.concatenate(
            [self._memory[:, 1:], r.values.reshape(-1, 1)], axis=1
        )

        # Look at indices that have been updated more than n_steps times
        ii = np.where(r.counts > self.n_steps)
        if len(ii) > 0:
            curr = self._memory[:, -1]
            saved = self._memory[:, 0]
            diffs = np.abs(curr[ii] - saved[ii])
            quots = np.divide(diffs, curr[ii], out=diffs, where=curr[ii] != 0)
            # quots holds the quotients when the denominator is non-zero, and
            # the absolute difference, which is just the memory, otherwise.
            if np.mean(quots) < self.rtol:
                self._converged = self.update_op(
                    self._converged, r.counts > self.n_steps
                )  # type: ignore
                if np.all(self._converged):
                    return Status.Converged
        return Status.Pending



import pytest
import ray
from ray.cluster_utils import Cluster

from pydvl.utils.config import ParallelConfig


@pytest.fixture(scope="module", params=["sequential", "ray-local", "ray-external"])
def parallel_config(request, num_workers):
    if request.param == "sequential":
        yield ParallelConfig(backend=request.param)
    elif request.param == "ray-local":
        yield ParallelConfig(backend="ray", n_cpus_local=num_workers)
        ray.shutdown()
    elif request.param == "ray-external":
        # Starts a head-node for the cluster.
        cluster = Cluster(
            initialize_head=True,
            head_node_args={
                "num_cpus": num_workers,
            },
        )
        yield ParallelConfig(backend="ray", address=cluster.address)
        ray.shutdown()
        cluster.shutdown()



import itertools
from typing import List, Tuple

import numpy as np
import pytest

from .conftest import create_mock_dataset

try:
    import torch.nn.functional as F
    from torch.optim import Adam, lr_scheduler

    from pydvl.influence.general import compute_influences
    from pydvl.influence.linear import (
        compute_linear_influences,
        influences_perturbation_linear_regression_analytical,
        influences_up_linear_regression_analytical,
    )
    from pydvl.influence.model_wrappers import TorchLinearRegression, TorchMLP
    from pydvl.utils.dataset import load_wine_dataset
except ImportError:
    pass


class InfluenceTestSettings:
    DATA_OUTPUT_NOISE: float = 0.01
    ACCEPTABLE_ABS_TOL_INFLUENCE: float = 5e-4
    ACCEPTABLE_ABS_TOL_INFLUENCE_CG: float = 1e-3

    INFLUENCE_TEST_CONDITION_NUMBERS: List[int] = [5]
    INFLUENCE_TRAINING_SET_SIZE: List[int] = [500]
    INFLUENCE_TEST_SET_SIZE: List[int] = [20]
    INFLUENCE_N_JOBS: List[int] = [1]
    INFLUENCE_DIMENSIONS: List[Tuple[int, int]] = [
        (10, 10),
        (20, 10),
        (3, 20),
        (20, 20),
    ]


test_cases = list(
    itertools.product(
        InfluenceTestSettings.INFLUENCE_TRAINING_SET_SIZE,
        InfluenceTestSettings.INFLUENCE_TEST_SET_SIZE,
        InfluenceTestSettings.INFLUENCE_DIMENSIONS,
        InfluenceTestSettings.INFLUENCE_TEST_CONDITION_NUMBERS,
        InfluenceTestSettings.INFLUENCE_N_JOBS,
    )
)


def lmb_test_case_to_str(packed_i_test_case):
    i, test_case = packed_i_test_case
    return (
        f"Problem #{i} of dimension {test_case[2]} with train size {test_case[0]}, "
        f"test size {test_case[1]}, condition number {test_case[3]} and {test_case[4]} jobs."
    )


test_case_ids = list(map(lmb_test_case_to_str, zip(range(len(test_cases)), test_cases)))


@pytest.mark.torch
@pytest.mark.parametrize(
    "train_set_size,test_set_size,problem_dimension,condition_number,n_jobs",
    test_cases,
    ids=test_case_ids,
)
def test_upweighting_influences_lr_analytical_cg(
    train_set_size: int,
    test_set_size: int,
    condition_number: float,
    linear_model: Tuple[np.ndarray, np.ndarray],
    n_jobs: int,
):
    A, _ = linear_model
    train_data, test_data = create_mock_dataset(
        linear_model, train_set_size, test_set_size
    )

    model = TorchLinearRegression(A.shape[0], A.shape[1], init=linear_model)
    loss = F.mse_loss

    influence_values_analytical = 2 * influences_up_linear_regression_analytical(
        linear_model,
        *train_data,
        *test_data,
    )

    influence_values = compute_influences(
        model,
        loss,
        *train_data,
        *test_data,
        progress=True,
        influence_type="up",
        inversion_method="cg",
        inversion_method_kwargs={"rtol": 10e-7},
    )
    assert np.logical_not(np.any(np.isnan(influence_values)))
    assert influence_values.shape == (len(test_data[0]), len(train_data[0]))
    influences_max_abs_diff = np.max(
        np.abs(influence_values - influence_values_analytical)
    )
    assert (
        influences_max_abs_diff < InfluenceTestSettings.ACCEPTABLE_ABS_TOL_INFLUENCE_CG
    ), "Upweighting influence values were wrong."


@pytest.mark.torch
@pytest.mark.parametrize(
    "train_set_size,test_set_size,problem_dimension,condition_number,n_jobs",
    test_cases,
    ids=test_case_ids,
)
def test_upweighting_influences_lr_analytical(
    train_set_size: int,
    test_set_size: int,
    condition_number: float,
    linear_model: Tuple[np.ndarray, np.ndarray],
    n_jobs: int,
):

    A, _ = tuple(linear_model)
    train_data, test_data = create_mock_dataset(
        linear_model, train_set_size, test_set_size
    )

    model = TorchLinearRegression(A.shape[0], A.shape[1], init=linear_model)
    loss = F.mse_loss

    influence_values_analytical = 2 * influences_up_linear_regression_analytical(
        linear_model,
        *train_data,
        *test_data,
    )

    influence_values = compute_influences(
        model,
        loss,
        *train_data,
        *test_data,
        progress=True,
        influence_type="up",
    )
    assert np.logical_not(np.any(np.isnan(influence_values)))
    assert influence_values.shape == (len(test_data[0]), len(train_data[0]))
    influences_max_abs_diff = np.max(
        np.abs(influence_values - influence_values_analytical)
    )
    assert (
        influences_max_abs_diff < InfluenceTestSettings.ACCEPTABLE_ABS_TOL_INFLUENCE
    ), "Upweighting influence values were wrong."


@pytest.mark.torch
@pytest.mark.parametrize(
    "train_set_size,test_set_size,problem_dimension,condition_number,n_jobs",
    test_cases,
    ids=test_case_ids,
)
def test_perturbation_influences_lr_analytical_cg(
    train_set_size: int,
    test_set_size: int,
    problem_dimension: int,
    condition_number: float,
    linear_model: Tuple[np.ndarray, np.ndarray],
    n_jobs: int,
):
    train_data, test_data = create_mock_dataset(
        linear_model, train_set_size, test_set_size
    )
    A, _ = linear_model

    model = TorchLinearRegression(A.shape[0], A.shape[1], init=linear_model)
    loss = F.mse_loss

    influence_values_analytical = (
        2
        * influences_perturbation_linear_regression_analytical(
            linear_model,
            *train_data,
            *test_data,
        )
    )
    influence_values = compute_influences(
        model,
        loss,
        *train_data,
        *test_data,
        progress=True,
        influence_type="perturbation",
        inversion_method="cg",
        inversion_method_kwargs={"rtol": 10e-7},
    )
    assert np.logical_not(np.any(np.isnan(influence_values)))
    assert influence_values.shape == (
        len(test_data[0]),
        len(train_data[0]),
        A.shape[1],
    )
    influences_max_abs_diff = np.max(
        np.abs(influence_values - influence_values_analytical)
    )
    assert (
        influences_max_abs_diff < InfluenceTestSettings.ACCEPTABLE_ABS_TOL_INFLUENCE
    ), "Perturbation influence values were wrong."


@pytest.mark.torch
@pytest.mark.parametrize(
    "train_set_size,test_set_size,problem_dimension,condition_number,n_jobs",
    test_cases,
    ids=test_case_ids,
)
def test_perturbation_influences_lr_analytical(
    train_set_size: int,
    test_set_size: int,
    problem_dimension: int,
    condition_number: float,
    linear_model: Tuple[np.ndarray, np.ndarray],
    n_jobs: int,
):
    train_data, test_data = create_mock_dataset(
        linear_model, train_set_size, test_set_size
    )
    A, _ = linear_model

    model = TorchLinearRegression(A.shape[0], A.shape[1], init=linear_model)
    loss = F.mse_loss

    influence_values_analytical = (
        2
        * influences_perturbation_linear_regression_analytical(
            linear_model,
            *train_data,
            *test_data,
        )
    )
    influence_values = compute_influences(
        model,
        loss,
        *train_data,
        *test_data,
        progress=True,
        influence_type="perturbation",
    )
    assert np.logical_not(np.any(np.isnan(influence_values)))
    assert influence_values.shape == (
        len(test_data[0]),
        len(train_data[0]),
        A.shape[1],
    )
    influences_max_abs_diff = np.max(
        np.abs(influence_values - influence_values_analytical)
    )
    assert (
        influences_max_abs_diff < InfluenceTestSettings.ACCEPTABLE_ABS_TOL_INFLUENCE
    ), "Perturbation influence values were wrong."


@pytest.mark.torch
@pytest.mark.parametrize(
    "train_set_size,test_set_size,problem_dimension,condition_number",
    itertools.product(
        InfluenceTestSettings.INFLUENCE_TRAINING_SET_SIZE,
        InfluenceTestSettings.INFLUENCE_TEST_SET_SIZE,
        InfluenceTestSettings.INFLUENCE_DIMENSIONS,
        InfluenceTestSettings.INFLUENCE_TEST_CONDITION_NUMBERS,
    ),
)
def test_linear_influences_up_perturbations_analytical(
    train_set_size: int,
    test_set_size: int,
    problem_dimension: int,
    condition_number: float,
    linear_model: Tuple[np.ndarray, np.ndarray],
):
    train_data, test_data = create_mock_dataset(
        linear_model, train_set_size, test_set_size
    )
    up_influences = compute_linear_influences(
        *train_data,
        *test_data,
        influence_type="up",
    )
    assert np.logical_not(np.any(np.isnan(up_influences)))
    assert up_influences.shape == (len(test_data[0]), len(train_data[0]))

    pert_influences = compute_linear_influences(
        *train_data,
        *test_data,
        influence_type="perturbation",
    )
    assert np.logical_not(np.any(np.isnan(pert_influences)))
    assert pert_influences.shape == (
        len(test_data[0]),
        len(train_data[0]),
        train_data[0].shape[1],
    )


@pytest.mark.torch
def test_influences_with_neural_network_explicit_hessian():
    train_ds, val_ds, test_ds, feature_names = load_wine_dataset(
        train_size=0.3, test_size=0.6
    )
    feature_dimension = train_ds[0].shape[1]
    unique_classes = np.unique(np.concatenate((train_ds[1], test_ds[1])))
    num_classes = len(unique_classes)
    num_epochs = 300
    network_size = [16, 16]
    nn = TorchMLP(feature_dimension, num_classes, network_size)
    optimizer = Adam(params=nn.parameters(), lr=0.001, weight_decay=0.001)
    loss = F.cross_entropy
    nn.fit(
        *train_ds,
        *test_ds,
        num_epochs=num_epochs,
        batch_size=32,
        loss=loss,
        optimizer=optimizer,
        scheduler=lr_scheduler.CosineAnnealingLR(optimizer, T_max=num_epochs),
    )

    model = nn
    loss = loss

    train_influences = compute_influences(
        model,
        loss,
        *train_ds,
        *test_ds,
        inversion_method="direct",
    )

    assert np.all(np.logical_not(np.isnan(train_influences)))



#!/usr/bin/env python3
"""
This script walks through the python source files and creates documentation in .rst format which can
then be compiled with Sphinx. It is suitable for a standard repository layout src/<library_name> as well as for
a repo containing multiple packages src/<package_1>, ...,  src/<package_n>.
"""
import argparse
import logging
import os
import shutil
from typing import Optional

log = logging.getLogger(__name__)


def module_template(module_qualname: str):
    module_name = module_qualname.split(".")[-1]
    title = module_name.replace("_", r"\_")
    template = f"""{title}
{"="*len(title)}

.. automodule:: {module_qualname}
   :members:
   :undoc-members:
   
   ----
 
.. footbibliography::

"""
    return template


def package_template(package_qualname: str, *, add_toctree: bool = True):
    package_name = package_qualname.split(".")[-1]
    title = package_name.replace("_", r"\_")
    template = f"""{title}
{"="*len(title)}

.. automodule:: {package_qualname}
   :members:
   :undoc-members:
"""
    if add_toctree:
        template += f"""
.. rubric:: Modules in this package

.. toctree::
   :glob:

   {package_name}/*

"""
    return template


def index_template(package_name: str, title: Optional[str] = None) -> str:
    if title is None:
        title = package_name.replace("_", r"\_")
    template = f"""{title}
{"="*len(title)}

.. automodule:: {package_name}
   :members:
   :undoc-members:

.. toctree::
   :glob:

   *
"""
    return template


def write_to_file(content: str, path: str):
    os.makedirs(os.path.dirname(path), exist_ok=True)
    with open(path, "w") as f:
        f.write(content)
    os.chmod(path, 0o666)


def make_rst(
    src_root: str = "src",
    docs_root: str = "docs",
    clean: bool = False,
    overwrite: bool = False,
    only_update: bool = True,
):
    """Creates / updates documentation in form of rst files for modules and
    packages. Does not delete any existing rst files if clean and overwrite are
    False. This method should be executed from the project's top-level
    directory.

    :param src_root: path to project's src directory that contains all packages,
        usually src. Most projects will only need one top-level package, then
        your layout typically should be src/<library_name>.
    :param docs_root: path to the project's docs directory containing the
        `conf.py` and the top level `index.rst`.
    :param clean: whether to completely clean the docs target directories
        beforehand, removing any existing files.
    :param overwrite: whether to overwrite existing rst files. This should be
        used with caution as it will delete all manual changes to documentation
        files.
    :param only_update: set to True if rst files should only be recreated if
        their modification date is earlier than that of the modules.
    :return:
    """
    docs_root = os.path.abspath(docs_root)
    src_root = os.path.abspath(src_root)

    for top_level_package_name in os.listdir(src_root):
        top_level_package_dir = os.path.join(src_root, top_level_package_name)
        # skipping things in src that are not packages, like .egg files
        if (
            not os.path.isdir(top_level_package_dir)
            or "." in top_level_package_name
            or top_level_package_name.startswith("_")
        ):
            continue

        log.info(
            f"Generating documentation for top-level package {top_level_package_name}"
        )
        top_level_package_docs_dir = os.path.join(docs_root, top_level_package_name)
        if clean and os.path.isdir(top_level_package_docs_dir):
            log.info(f"Deleting {top_level_package_docs_dir} since clean=True")
            shutil.rmtree(top_level_package_docs_dir)

        index_rst_path = os.path.join(docs_root, top_level_package_name, "index.rst")
        log.info(f"Creating {index_rst_path}")
        write_to_file(
            index_template(top_level_package_name, "API Reference"), index_rst_path
        )

        for root, dirnames, filenames in os.walk(top_level_package_dir):
            if os.path.basename(root).startswith("_"):
                log.debug(f"Skipping doc generation in {root}")
                continue

            base_package_relpath = os.path.relpath(root, start=top_level_package_dir)
            base_package_qualname = os.path.relpath(root, start=src_root).replace(
                os.path.sep, "."
            )

            for dirname in dirnames:
                if not dirname.startswith("_"):
                    package_qualname = f"{base_package_qualname}.{dirname}"
                    package_rst_path = os.path.abspath(
                        os.path.join(
                            top_level_package_docs_dir,
                            base_package_relpath,
                            f"{dirname}.rst",
                        )
                    )
                    package_path = os.path.join(root, dirname)
                    add_toctree = True
                    package_dir_content = list(
                        filter(lambda x: x != "__pycache__", os.listdir(package_path))
                    )
                    if package_dir_content == ["__init__.py"]:
                        add_toctree = False

                    try:
                        dir_path = os.path.join(root, dirname)
                        if only_update and os.path.getmtime(
                            dir_path
                        ) <= os.path.getmtime(package_rst_path):
                            log.info(
                                f"Package {dir_path} hasn't been modified, skipping."
                            )
                            continue
                    except FileNotFoundError:
                        pass

                    log.info(f"Writing package documentation to {package_rst_path}")
                    write_to_file(
                        package_template(package_qualname, add_toctree=add_toctree),
                        package_rst_path,
                    )

            for filename in filenames:
                base_name, ext = os.path.splitext(filename)
                if ext == ".py" and not filename.startswith("_"):
                    module_qualname = f"{base_package_qualname}.{filename[:-3]}"
                    module_rst_path = os.path.abspath(
                        os.path.join(
                            top_level_package_docs_dir,
                            base_package_relpath,
                            f"{base_name}.rst",
                        )
                    )
                    if os.path.exists(module_rst_path) and not overwrite:
                        log.debug(f"{module_rst_path} already exists, skipping it")

                    try:
                        file_path = os.path.join(root, filename)
                        if only_update and os.path.getmtime(
                            file_path
                        ) <= os.path.getmtime(module_rst_path):
                            log.info(
                                f"Module {file_path} hasn't been modified, skipping."
                            )
                            continue
                    except FileNotFoundError:
                        pass

                    log.info(f"Writing module documentation to {module_rst_path}")
                    write_to_file(module_template(module_qualname), module_rst_path)


if __name__ == "__main__":
    parser = argparse.ArgumentParser(
        description="A tool to create RST files for all source files in the library",
        formatter_class=argparse.ArgumentDefaultsHelpFormatter,
    )
    parser.add_argument(
        "-s", "--source", help="Root of the sources", type=str, default="src"
    )

    parser.add_argument(
        "-d", "--doc", help="Root of the documentation", type=str, default="docs"
    )

    parser.add_argument(
        "-u",
        "--update",
        help="Whether to only update rst files if sources are newer",
        action="store_true",
    )
    parser.add_argument(
        "-c", "--clean", help="Wipe docs before starting", action="store_true"
    )
    args = parser.parse_args()

    logging.basicConfig(level=logging.INFO)
    make_rst(
        src_root=args.source,
        docs_root=args.doc,
        clean=args.clean,
        only_update=args.update,
    )



from pydvl.utils import Status


def test_or_status():
    """The result of bitwise or-ing two valuation statuses is given by the
    following table:

        |   | P | C | F |
        |---|---|---|---|
        | P | P | C | P |
        | C | C | C | C |
        | F | P | C | F |

    where P = Pending, C = Converged, F = Failed.
    """
    assert Status.Pending | Status.Pending == Status.Pending
    assert Status.Pending | Status.Converged == Status.Converged
    assert Status.Pending | Status.Failed == Status.Pending
    assert Status.Converged | Status.Pending == Status.Converged
    assert Status.Converged | Status.Converged == Status.Converged
    assert Status.Converged | Status.Failed == Status.Converged
    assert Status.Failed | Status.Pending == Status.Pending
    assert Status.Failed | Status.Converged == Status.Converged
    assert Status.Failed | Status.Failed == Status.Failed


def test_and_status():
    """The result of bitwise &-ing two valuation statuses is given by the
    following table:

        |   | P | C | F |
        |---|---|---|---|
        | P | P | P | F |
        | C | P | C | F |
        | F | F | F | F |

    where P = Pending, C = Converged, F = Failed.
    """
    assert Status.Pending & Status.Pending == Status.Pending
    assert Status.Pending & Status.Converged == Status.Pending
    assert Status.Pending & Status.Failed == Status.Failed
    assert Status.Converged & Status.Pending == Status.Pending
    assert Status.Converged & Status.Converged == Status.Converged
    assert Status.Converged & Status.Failed == Status.Failed
    assert Status.Failed & Status.Pending == Status.Failed
    assert Status.Failed & Status.Converged == Status.Failed
    assert Status.Failed & Status.Failed == Status.Failed


def test_not_status():
    """The result of bitwise negation of a Status is ``Failed``
    if the status is ``Converged``, or ``Converged`` otherwise:

        ``~P == C, ~C == F, ~F == C``
    """
    assert ~Status.Pending == Status.Converged
    assert ~Status.Converged == Status.Failed
    assert ~Status.Failed == Status.Converged



import logging

import pytest
from numpy.testing import assert_almost_equal

from pydvl.value.least_core import montecarlo_least_core
from tests.value import check_values

logger = logging.getLogger(__name__)


@pytest.mark.parametrize(
    "test_utility, rtol, n_iterations",
    [
        (("miner", {"n_miners": 8}), 0.1, 128),
        (("gloves", {"left": 10, "right": 5}), 0.2, 10000),
    ],
    indirect=["test_utility"],
)
@pytest.mark.parametrize("n_jobs", [1, -1])
@pytest.mark.parametrize("non_negative_subsidy", (True, False))
def test_montecarlo_least_core(
    test_utility, rtol, n_iterations, n_jobs, non_negative_subsidy
):
    u, exact_values = test_utility

    values = montecarlo_least_core(
        u,
        n_iterations=n_iterations,
        non_negative_subsidy=non_negative_subsidy,
        progress=False,
        n_jobs=n_jobs,
    )
    if non_negative_subsidy:
        check_values(values, exact_values)
        # Sometimes the subsidy is negative but really close to zero
        # Due to numerical errors
        if values.subsidy < 0:
            assert_almost_equal(values.subsidy, 0.0, decimal=5)
    else:
        check_values(values, exact_values, extra_values_names=["subsidy"])



"""
This module contains all functions for the closed form computation of influences
for standard linear regression.
"""
from typing import TYPE_CHECKING, Tuple

import numpy as np
from sklearn.linear_model import LinearRegression

from ..utils.numeric import (
    linear_regression_analytical_derivative_d2_theta,
    linear_regression_analytical_derivative_d_theta,
    linear_regression_analytical_derivative_d_x_d_theta,
)
from .general import InfluenceType

if TYPE_CHECKING:
    from numpy.typing import NDArray

__all__ = ["compute_linear_influences"]


def compute_linear_influences(
    x: "NDArray",
    y: "NDArray",
    x_test: "NDArray",
    y_test: "NDArray",
    influence_type: InfluenceType = InfluenceType.Up,
):
    """Calculate the influence each training sample on the loss computed over a
    validation set for an ordinary least squares model ($y = A x + b$ with
    quadratic loss).

    :param x: An array of shape (M, K) containing the features of training data.
    :param y: An array of shape (M, L) containing the targets of training data.
    :param x_test: An array of shape (N, K) containing the features of the
        test set.
    :param y_test: An array of shape (N, L) containing the targets of the test
        set.
    :param influence_type: Which algorithm to use to calculate influences.
        Currently supported options: 'up' or 'perturbation'.
    :returns: An array of shape (B, C) with the influences of the training
        points on the test data.
    """

    lr = LinearRegression()
    lr.fit(x, y)
    A = lr.coef_
    b = lr.intercept_

    if influence_type not in list(InfluenceType):
        raise NotImplementedError(
            f"Only up-weighting and perturbation influences are supported, but got {influence_type=}"
        )

    if influence_type == InfluenceType.Up:
        return influences_up_linear_regression_analytical(
            (A, b),
            x,
            y,
            x_test,
            y_test,
        )
    elif influence_type == InfluenceType.Perturbation:
        return influences_perturbation_linear_regression_analytical(
            (A, b),
            x,
            y,
            x_test,
            y_test,
        )


def influences_up_linear_regression_analytical(
    linear_model: Tuple["NDArray", "NDArray"],
    x: "NDArray",
    y: "NDArray",
    x_test: "NDArray",
    y_test: "NDArray",
):
    """Calculate the influence each training sample on the loss computed over a
     validation set for an ordinary least squares model (Ax+b=y with quadratic
     loss).

    This method uses the

    :param linear_model: A tuple of arrays of shapes (N, M) and N representing A
        and b respectively.
    :param x: An array of shape (M, K) containing the features of the
        training set.
    :param y: An array of shape (M, L) containing the targets of the
        training set.
    :param x_test: An array of shape (N, K) containing the features of the test
        set.
    :param y_test: An array of shape (N, L) containing the targets of the test
        set.
    :returns: An array of shape (B, C) with the influences of the training points
        on the test points.
    """

    test_grads_analytical = linear_regression_analytical_derivative_d_theta(
        linear_model,
        x_test,
        y_test,
    )
    train_grads_analytical = linear_regression_analytical_derivative_d_theta(
        linear_model,
        x,
        y,
    )
    hessian_analytical = linear_regression_analytical_derivative_d2_theta(
        linear_model,
        x,
        y,
    )
    s_test_analytical = np.linalg.solve(hessian_analytical, test_grads_analytical.T).T
    result: "NDArray" = np.einsum(
        "ia,ja->ij", s_test_analytical, train_grads_analytical
    )
    return result


def influences_perturbation_linear_regression_analytical(
    linear_model: Tuple["NDArray", "NDArray"],
    x: "NDArray",
    y: "NDArray",
    x_test: "NDArray",
    y_test: "NDArray",
):
    """Calculate the influences of each training sample onto the
    validation set for a linear model Ax+b=y.

    :param linear_model: A tuple of np.ndarray' of shape (N, M) and (N)
        representing A and b respectively.
    :param x: An array of shape (M, K) containing the features of the
        input data.
    :param y: An array of shape (M, L) containing the targets of the input
        data.
    :param x_test: An array of shape (N, K) containing the features of the test
        set.
    :param y_test: An array of shape (N, L) containing the targets of the test
        set.
    :returns: An array of shape (B, C, M) with the influences of the training
        points on the test points for each feature.
    """

    test_grads_analytical = linear_regression_analytical_derivative_d_theta(
        linear_model,
        x_test,
        y_test,
    )
    train_second_deriv_analytical = linear_regression_analytical_derivative_d_x_d_theta(
        linear_model,
        x,
        y,
    )

    hessian_analytical = linear_regression_analytical_derivative_d2_theta(
        linear_model,
        x,
        y,
    )
    s_test_analytical = np.linalg.solve(hessian_analytical, test_grads_analytical.T).T
    result: "NDArray" = np.einsum(
        "ia,jab->ijb", s_test_analytical, train_second_deriv_analytical
    )
    return result



from typing import TYPE_CHECKING, Tuple

import numpy as np
import pytest
from sklearn.preprocessing import MinMaxScaler

from pydvl.utils import Dataset, random_matrix_with_condition_number

if TYPE_CHECKING:
    from numpy.typing import NDArray


@pytest.fixture
def input_dimension(request) -> int:
    return request.param


@pytest.fixture
def output_dimension(request) -> int:
    return request.param


@pytest.fixture
def problem_dimension(request) -> int:
    return request.param


@pytest.fixture
def batch_size(request) -> int:
    return request.param


@pytest.fixture
def condition_number(request) -> float:
    return request.param


@pytest.fixture(scope="function")
def quadratic_linear_equation_system(quadratic_matrix: np.ndarray, batch_size: int):
    A = quadratic_matrix
    problem_dimension = A.shape[0]
    b = np.random.random([batch_size, problem_dimension])
    return A, b


@pytest.fixture(scope="function")
def quadratic_matrix(problem_dimension: int, condition_number: float):
    return random_matrix_with_condition_number(problem_dimension, condition_number)


@pytest.fixture(scope="function")
def singular_quadratic_linear_equation_system(
    quadratic_matrix: np.ndarray, batch_size: int
):
    A = quadratic_matrix
    problem_dimension = A.shape[0]
    i, j = tuple(np.random.choice(problem_dimension, replace=False, size=2))
    if j < i:
        i, j = j, i

    v = (A[i] + A[j]) / 2
    A[i], A[j] = v, v
    b = np.random.random([batch_size, problem_dimension])
    return A, b


@pytest.fixture(scope="function")
def linear_model(problem_dimension: Tuple[int, int], condition_number: float):
    output_dimension, input_dimension = problem_dimension
    A = random_matrix_with_condition_number(
        max(input_dimension, output_dimension), condition_number
    )
    A = A[:output_dimension, :input_dimension]
    b = np.random.uniform(size=[output_dimension])
    return A, b


def create_mock_dataset(
    linear_model: Tuple["NDArray[np.float_]", "NDArray[np.float_]"],
    train_set_size: int,
    test_set_size: int,
    noise: float = 0.01,
) -> Tuple[
    Tuple["NDArray[np.float_]", "NDArray[np.float_]"],
    Tuple["NDArray[np.float_]", "NDArray[np.float_]"],
]:
    A, b = linear_model
    o_d, i_d = tuple(A.shape)
    data_model = lambda x: np.random.normal(x @ A.T + b, noise)

    x_train = np.random.uniform(size=[train_set_size, i_d])
    y_train = data_model(x_train)
    x_test = np.random.uniform(size=[test_set_size, i_d])
    y_test = data_model(x_test)
    dataset = Dataset(
        x_train=x_train,
        y_train=y_train,
        x_test=x_test,
        y_test=y_test,
        is_multi_output=True,
    )

    scaler_x = MinMaxScaler()
    scaler_y = MinMaxScaler()
    dataset.x_train = scaler_x.fit_transform(dataset.x_train)
    dataset.y_train = scaler_y.fit_transform(dataset.y_train)
    dataset.x_test = scaler_x.transform(dataset.x_test)
    dataset.y_test = scaler_y.transform(dataset.y_test)
    return (x_train, y_train), (x_test, y_test)



import os
from abc import ABCMeta, abstractmethod
from dataclasses import asdict
from typing import (
    Any,
    Callable,
    Dict,
    Iterable,
    List,
    Optional,
    Tuple,
    Type,
    TypeVar,
    Union,
)

import ray
from ray import ObjectRef

from ..config import ParallelConfig

__all__ = ["init_parallel_backend", "effective_n_jobs", "available_cpus"]

T = TypeVar("T")

_PARALLEL_BACKENDS: Dict[str, "Type[BaseParallelBackend]"] = {}


class NoPublicConstructor(ABCMeta):
    """Metaclass that ensures a private constructor

    If a class uses this metaclass like this:

        class SomeClass(metaclass=NoPublicConstructor):
            pass

    If you try to instantiate your class (`SomeClass()`),
    a `TypeError` will be thrown.

    Taken almost verbatim from:
    https://stackoverflow.com/a/64682734
    """

    def __call__(cls, *args, **kwargs):
        raise TypeError(
            f"{cls.__module__}.{cls.__qualname__} cannot be initialized directly. "
            "Use init_parallel_backend() instead."
        )

    def _create(cls, *args: Any, **kwargs: Any):
        return super().__call__(*args, **kwargs)


class BaseParallelBackend(metaclass=NoPublicConstructor):
    """Abstract base class for all parallel backends"""

    config: Dict[str, Any] = {}

    def __init_subclass__(cls, *, backend_name: str, **kwargs):
        global _PARALLEL_BACKENDS
        _PARALLEL_BACKENDS[backend_name] = cls
        super().__init_subclass__(**kwargs)

    @abstractmethod
    def get(self, v: Any, *args, **kwargs):
        ...

    @abstractmethod
    def put(self, v: Any, *args, **kwargs) -> Any:
        ...

    @abstractmethod
    def wrap(self, fun: Callable, **kwargs) -> Callable:
        ...

    @abstractmethod
    def wait(self, v: Any, *args, **kwargs) -> Any:
        ...

    @abstractmethod
    def _effective_n_jobs(self, n_jobs: int) -> int:
        ...

    def effective_n_jobs(self, n_jobs: int = -1) -> int:
        if n_jobs == 0:
            raise ValueError("n_jobs == 0 in Parallel has no meaning")
        n_jobs = self._effective_n_jobs(n_jobs)
        return n_jobs

    def __repr__(self) -> str:
        return f"<{self.__class__.__name__}: {self.config}>"


class SequentialParallelBackend(BaseParallelBackend, backend_name="sequential"):
    """Class used to run jobs sequentially and locally.

    It shouldn't be initialized directly. You should instead call
    :func:`~pydvl.utils.parallel.backend.init_parallel_backend`.

    :param config: instance of :class:`~pydvl.utils.config.ParallelConfig` with number of cpus
    """

    def __init__(self, config: ParallelConfig):
        self.config = {}

    def get(self, v: Any, *args, **kwargs):
        return v

    def put(self, v: Any, *args, **kwargs) -> Any:
        return v

    def wrap(self, fun: Callable, **kwargs) -> Callable:
        """Wraps a function for sequential execution.

        This is a noop and kwargs are ignored."""
        return fun

    def wait(self, v: Any, *args, **kwargs) -> Tuple[list, list]:
        return v, []

    def _effective_n_jobs(self, n_jobs: int) -> int:
        return 1


class RayParallelBackend(BaseParallelBackend, backend_name="ray"):
    """Class used to wrap ray to make it transparent to algorithms.

    It shouldn't be initialized directly. You should instead call
    :func:`~pydvl.utils.parallel.backend.init_parallel_backend`.

    :param config: instance of :class:`~pydvl.utils.config.ParallelConfig` with
        cluster address, number of cpus, etc.
    """

    def __init__(self, config: ParallelConfig):
        config_dict = asdict(config)
        config_dict.pop("backend")
        n_cpus_local = config_dict.pop("n_cpus_local")
        if config_dict.get("address", None) is None:
            config_dict["num_cpus"] = n_cpus_local
        self.config = config_dict
        if not ray.is_initialized():
            ray.init(**self.config)

    def get(
        self,
        v: Union[ObjectRef, Iterable[ObjectRef], T],
        *args,
        **kwargs,
    ) -> Union[T, Any]:
        timeout: Optional[float] = kwargs.get("timeout", None)
        if isinstance(v, ObjectRef):
            return ray.get(v, timeout=timeout)
        elif isinstance(v, Iterable):
            return [self.get(x, timeout=timeout) for x in v]
        else:
            return v

    def put(self, v: T, *args, **kwargs) -> Union["ObjectRef[T]", T]:
        try:
            return ray.put(v, **kwargs)  # type: ignore
        except TypeError:
            return v  # type: ignore

    def wrap(self, fun: Callable, **kwargs) -> Callable:
        """Wraps a function as a ray remote.

        :param fun: the function to wrap
        :param kwargs: keyword arguments to pass to @ray.remote

        :return: The `.remote` method of the ray `RemoteFunction`.
        """
        if len(kwargs) > 0:
            return ray.remote(**kwargs)(fun).remote  # type: ignore
        return ray.remote(fun).remote  # type: ignore

    def wait(
        self,
        v: List["ObjectRef"],
        *args,
        **kwargs,
    ) -> Tuple[List[ObjectRef], List[ObjectRef]]:
        num_returns: int = kwargs.get("num_returns", 1)
        timeout: Optional[float] = kwargs.get("timeout", None)
        return ray.wait(  # type: ignore
            v,
            num_returns=num_returns,
            timeout=timeout,
        )

    def _effective_n_jobs(self, n_jobs: int) -> int:
        if n_jobs < 0:
            ray_cpus = int(ray._private.state.cluster_resources()["CPU"])  # type: ignore
            eff_n_jobs = ray_cpus
        else:
            eff_n_jobs = n_jobs
        return eff_n_jobs


def init_parallel_backend(
    config: ParallelConfig,
) -> BaseParallelBackend:
    """Initializes the parallel backend and returns an instance of it.

    :param config: instance of :class:`~pydvl.utils.config.ParallelConfig`
        with cluster address, number of cpus, etc.

    :Example:

    >>> from pydvl.utils.parallel.backend import init_parallel_backend
    >>> from pydvl.utils.config import ParallelConfig
    >>> config = ParallelConfig(backend="ray")
    >>> parallel_backend = init_parallel_backend(config)
    >>> parallel_backend
    <RayParallelBackend: {'address': None, 'logging_level': 30, 'num_cpus': None}>

    """
    try:
        parallel_backend_cls = _PARALLEL_BACKENDS[config.backend]
    except KeyError:
        raise NotImplementedError(f"Unexpected parallel backend {config.backend}")
    parallel_backend = parallel_backend_cls._create(config)
    return parallel_backend  # type: ignore


def available_cpus() -> int:
    """Platform-independent count of available cores.

    FIXME: do we really need this or is `os.cpu_count` enough? Is this portable?
    :return: Number of cores, or 1 if it is not possible to determine.
    """
    from platform import system

    if system() != "Linux":
        return os.cpu_count() or 1
    return len(os.sched_getaffinity(0))


def effective_n_jobs(n_jobs: int, config: ParallelConfig = ParallelConfig()) -> int:
    """Returns the effective number of jobs.

    This number may vary depending on the parallel backend and the resources
    available.

    :param n_jobs: the number of jobs requested. If -1, the number of available
        CPUs is returned.
    :param config: instance of :class:`~pydvl.utils.config.ParallelConfig` with
        cluster address, number of cpus, etc.
    :return: the effective number of jobs, guaranteed to be >= 1.
    :raises RuntimeError: if the effective number of jobs returned by the backend
        is < 1.
    """
    parallel_backend = init_parallel_backend(config)
    if (eff_n_jobs := parallel_backend.effective_n_jobs(n_jobs)) < 1:
        raise RuntimeError(
            f"Invalid number of jobs {eff_n_jobs} obtained from parallel backend {config.backend}"
        )
    return eff_n_jobs



import logging
from abc import ABC, abstractmethod
from typing import TYPE_CHECKING, Any, Callable, List, Optional, Tuple, Union

import numpy as np

from ...utils import maybe_progress

try:
    import torch
    import torch.nn as nn
    from torch.nn import Softmax, Tanh
    from torch.optim import Optimizer
    from torch.optim.lr_scheduler import _LRScheduler
    from torch.utils.data import DataLoader, TensorDataset

    _TORCH_INSTALLED = True
except ImportError:
    _TORCH_INSTALLED = False

if TYPE_CHECKING:
    from numpy.typing import NDArray

__all__ = [
    "TorchLinearRegression",
    "TorchBinaryLogisticRegression",
    "TorchMLP",
    "TorchModel",
]

logger = logging.getLogger(__name__)


class TorchModelBase(ABC):
    def __init__(self):
        if not _TORCH_INSTALLED:
            raise RuntimeWarning("This function requires PyTorch.")

    @abstractmethod
    def forward(self, x: torch.Tensor) -> torch.Tensor:
        pass

    def fit(
        self,
        x_train: Union["NDArray[np.float_]", torch.tensor],
        y_train: Union["NDArray[np.float_]", torch.tensor],
        x_val: Union["NDArray[np.float_]", torch.tensor],
        y_val: Union["NDArray[np.float_]", torch.tensor],
        loss: Callable[[torch.Tensor, torch.Tensor], torch.Tensor],
        optimizer: Optimizer,
        scheduler: Optional[_LRScheduler] = None,
        num_epochs: int = 1,
        batch_size: int = 64,
        progress: bool = True,
    ) -> Tuple["NDArray[np.float_]", "NDArray[np.float_]"]:
        """
        Wrapper of pytorch fit method. It fits the model to the supplied data.
        It represents a simple machine learning loop, iterating over a number of
        epochs, sampling data with a certain batch size, calculating gradients and updating the parameters through a
        loss function.
        :param x: Matrix of shape [NxD] representing the features x_i.
        :param y: Matrix of shape [NxK] representing the prediction targets y_i.
        :param optimizer: Select either ADAM or ADAM_W.
        :param scheduler: A pytorch scheduler. If None, no scheduler is used.
        :param num_epochs: Number of epochs to repeat training.
        :param batch_size: Batch size to use in training.
        :param progress: True, iff progress shall be printed.
        :param tensor_type: accuracy of tensors. Typically 'float' or 'long'
        """
        x_train = torch.as_tensor(x_train).clone()
        y_train = torch.as_tensor(y_train).clone()
        x_val = torch.as_tensor(x_val).clone()
        y_val = torch.as_tensor(y_val).clone()

        dataset = TensorDataset(x_train, y_train)
        dataloader = DataLoader(dataset, batch_size=batch_size)
        train_loss = []
        val_loss = []

        for epoch in maybe_progress(range(num_epochs), progress, desc="Model fitting"):
            batch_loss = []
            for train_batch in dataloader:
                batch_x, batch_y = train_batch
                pred_y = self.forward(batch_x)
                loss_value = loss(torch.squeeze(pred_y), torch.squeeze(batch_y))
                batch_loss.append(loss_value.item())

                logger.debug(f"Epoch: {epoch} ---> Training loss: {loss_value.item()}")
                loss_value.backward()
                optimizer.step()
                optimizer.zero_grad()

                if scheduler:
                    scheduler.step()
            pred_val = self.forward(x_val)
            epoch_val_loss = loss(torch.squeeze(pred_val), torch.squeeze(y_val)).item()
            mean_epoch_train_loss = np.mean(batch_loss)
            val_loss.append(epoch_val_loss)
            train_loss.append(mean_epoch_train_loss)
            logger.info(
                f"Epoch: {epoch} ---> Training loss: {mean_epoch_train_loss}, Validation loss: {epoch_val_loss}"
            )
        return np.array(train_loss), np.array(val_loss)

    def predict(self, x: torch.Tensor) -> "NDArray[np.float_]":
        """
        Use internal model to deliver prediction in numpy.
        :param x: A np.ndarray [NxD] representing the features x_i.
        :returns: A np.ndarray [NxK] representing the predicted values.
        """
        return self.forward(x).detach().numpy()  # type: ignore

    def score(
        self,
        x: torch.Tensor,
        y: torch.Tensor,
        score: Callable[[torch.Tensor, torch.Tensor, Any], torch.Tensor],
    ) -> float:
        """
        Use internal model to measure how good is prediction through a loss function.
        :param x: A np.ndarray [NxD] representing the features x_i.
        :param y: A np.ndarray [NxK] representing the predicted target values y_i.
        :returns: The aggregated value over all samples N.
        """
        return score(self.forward(x), y).detach().numpy()  # type: ignore


class TorchModel(TorchModelBase):
    def __init__(self, model: nn.Module):
        self.model = model

    def forward(self, x: torch.Tensor) -> torch.Tensor:
        return self.model(x)

    def __call__(self, *args, **kwargs):
        return self.model(*args, **kwargs)


class TorchLinearRegression(nn.Module, TorchModelBase):
    """
    A simple linear regression model (with bias) f(x)=Ax+b.
    """

    def __init__(
        self,
        n_input: int,
        n_output: int,
        init: Tuple["NDArray[np.float_]", "NDArray[np.float_]"] = None,
    ):
        """
        :param n_input: input to the model.
        :param n_output: output of the model
        :param init A tuple with two matrices, namely A of shape [K, D] and b of shape [K]. If set to None Xavier
        uniform initialization is used.
        """
        super().__init__()
        self.n_input = n_input
        self.n_output = n_output
        if init is None:
            r = np.sqrt(6 / (n_input + n_output))
            init_A = np.random.uniform(-r, r, size=[n_output, n_input])
            init_b = np.zeros(n_output)
            init = (init_A, init_b)

        self.A = nn.Parameter(
            torch.tensor(init[0], dtype=torch.float64), requires_grad=True
        )
        self.b = nn.Parameter(
            torch.tensor(init[1], dtype=torch.float64), requires_grad=True
        )

    def forward(self, x: torch.Tensor) -> torch.Tensor:
        """
        Calculate A @ x + b using RAM-optimized calculation layout.
        :param x: Tensor [NxD] representing the features x_i.
        :returns A tensor [NxK] representing the outputs y_i.
        """
        return x @ self.A.T + self.b


class TorchBinaryLogisticRegression(nn.Module, TorchModelBase):
    """
    A simple binary logistic regression model p(y)=sigmoid(dot(a, x) + b).
    """

    def __init__(
        self,
        n_input: int,
        init: Tuple["NDArray[np.float_]", "NDArray[np.float_]"] = None,
    ):
        """
        :param n_input: Number of feature inputs to the BinaryLogisticRegressionModel.
        :param init A tuple representing the initialization for the weight matrix A and the bias b. If set to None
        sample the values uniformly using the Xavier rule.
        """
        super().__init__()
        self.n_input = n_input
        if init is None:
            init_A = np.random.normal(0, 0.02, size=(1, n_input))
            init_b = np.random.normal(0, 0.02, size=(1))
            init = (init_A, init_b)

        self.A = nn.Parameter(torch.tensor(init[0]), requires_grad=True)
        self.b = nn.Parameter(torch.tensor(init[1]), requires_grad=True)

    def forward(self, x: Union["NDArray[np.float_]", torch.Tensor]) -> torch.Tensor:
        """
        Calculate sigmoid(dot(a, x) + b) using RAM-optimized calculation layout.
        :param x: Tensor [NxD] representing the features x_i.
        :returns: A tensor [N] representing the probabilities for p(y_i).
        """
        x = torch.as_tensor(x)
        return torch.sigmoid(x @ self.A.T + self.b)


class TorchMLP(nn.Module, TorchModelBase):
    """
    A simple fully-connected neural network f(x) model defined by y = v_K, v_i = o(A v_(i-1) + b), v_1 = x. It contains
    K layers and K - 2 hidden layers. It holds that K >= 2, because every network contains a input and output.
    """

    def __init__(
        self,
        n_input: int,
        n_output: int,
        n_neurons_per_layer: List[int],
        output_probabilities: bool = True,
        init: List[Tuple["NDArray[np.float_]", "NDArray[np.float_]"]] = None,
    ):
        """
        :param n_input: Number of feature in input.
        :param n_output: Output length.
        :param n_neurons_per_layer: Each integer represents the size of a hidden layer. Overall this list has K - 2
        :param output_probabilities: True, if the model should output probabilities. In the case of n_output 2 the
        number of outputs reduce to 1.
        :param init: A list of tuple of np.ndarray representing the internal weights.
        """
        super().__init__()
        self.n_input = n_input
        self.n_output = 1 if output_probabilities and n_output == 2 else n_output

        self.n_hidden_layers = n_neurons_per_layer
        self.output_probabilities = output_probabilities

        all_dimensions = [self.n_input] + self.n_hidden_layers + [self.n_output]
        layers = []
        num_layers = len(all_dimensions) - 1
        for num_layer, (in_features, out_features) in enumerate(
            zip(all_dimensions[:-1], all_dimensions[1:])
        ):
            linear_layer = nn.Linear(
                in_features, out_features, bias=num_layer < len(all_dimensions) - 2
            )

            if init is None:
                torch.nn.init.xavier_uniform_(linear_layer.weight)
                if num_layer < len(all_dimensions) - 2:
                    linear_layer.bias.data.fill_(0.01)

            else:
                A, b = init[num_layer]
                linear_layer.weight.data = A
                if num_layer < len(all_dimensions) - 2:
                    linear_layer.bias.data = b

            layers.append(linear_layer)
            if num_layer < num_layers - 1:
                layers.append(Tanh())
            elif self.output_probabilities:
                layers.append(Softmax(dim=-1))

        self.layers = nn.Sequential(*layers)

    def forward(self, x: torch.Tensor) -> torch.Tensor:
        """
        Perform forward-pass through the network.
        :param x: Tensor input of shape [NxD].
        :returns: Tensor output of shape[NxK].
        """
        return self.layers(x)



import os
import pickle as pkl
from copy import deepcopy
from pathlib import Path
from typing import TYPE_CHECKING, Any, Dict, Iterable, List, Optional, Sequence, Tuple

import matplotlib.patches as mpatches
import matplotlib.pyplot as plt
import numpy as np
import pandas as pd
import torch.nn as nn
from PIL.JpegImagePlugin import JpegImageFile
from torch.optim import Adam
from torchvision.models import ResNet18_Weights, resnet18

from pydvl.influence.model_wrappers import TorchModel
from pydvl.utils import Dataset

try:
    import torch

    _TORCH_INSTALLED = True
except ImportError:
    _TORCH_INSTALLED = False

if TYPE_CHECKING:
    from numpy.typing import NDArray

MODEL_PATH = Path().resolve().parent / "data" / "models"


def new_resnet_model(output_size: int) -> TorchModel:
    model = resnet18(weights=ResNet18_Weights.IMAGENET1K_V1)

    for param in model.parameters():
        param.requires_grad = False

    # Fine-tune final few layers
    model.avgpool = nn.AdaptiveAvgPool2d(1)
    n_features = model.fc.in_features
    model.fc = nn.Linear(n_features, output_size)
    device = torch.device("cuda:0" if torch.cuda.is_available() else "cpu")
    model.to(device)

    return TorchModel(model)


class TrainingManager:
    """A simple class to handle persistence of the model for the notebook
    `influence_imagenet.ipynb`
    """

    def __init__(
        self,
        name: str,
        model: TorchModel,
        loss: torch.nn.modules.loss._Loss,
        train_x: "NDArray[np.float_]",
        train_y: "NDArray[np.float_]",
        val_x: "NDArray[np.float_]",
        val_y: "NDArray[np.float_]",
        data_dir: Path,
    ):
        self.name = name
        self.model_wrapper = model
        self.loss = loss
        self.train_x, self.train_y = train_x, train_y
        self.val_x, self.val_y = val_x, val_y
        self.data_dir = data_dir
        os.makedirs(self.data_dir, exist_ok=True)

    @property
    def model(self) -> nn.Module:
        return self.model_wrapper.model

    def train(
        self,
        n_epochs: int,
        lr: float = 0.001,
        batch_size: int = 1000,
        use_cache: bool = True,
    ) -> Tuple["NDArray[np.float_]", "NDArray[np.float_]"]:
        """
        :return: Tuple of training_loss, validation_loss
        """
        if use_cache:
            try:
                training_loss, validation_loss = self.load()
                print("Cached model found, loading...")
                return training_loss, validation_loss
            except:
                print(f"No pretrained model found. Training for {n_epochs} epochs:")

        optimizer = Adam(self.model.parameters(), lr=lr)

        training_loss, validation_loss = self.model_wrapper.fit(
            x_train=self.train_x,
            y_train=self.train_y,
            x_val=self.val_x,
            y_val=self.val_y,
            loss=self.loss,
            optimizer=optimizer,
            num_epochs=n_epochs,
            batch_size=batch_size,
        )
        if use_cache:
            self.save(training_loss, validation_loss)

        return training_loss, validation_loss

    def save(
        self, training_loss: "NDArray[np.float_]", validation_loss: "NDArray[np.float_]"
    ):
        """Saves the model weights and training and validation losses.

        :param training_loss: list of training losses, one per epoch
        :param validation_loss: list of validation losses, also one per epoch
        """
        torch.save(self.model.state_dict(), self.data_dir / f"{self.name}_weights.pth")
        with open(self.data_dir / f"{self.name}_train_val_loss.pkl", "wb") as file:
            pkl.dump([training_loss, validation_loss], file)

    def load(self) -> Tuple["NDArray[np.float_]", "NDArray[np.float_]"]:
        """Loads model weights and training and validation losses.
        :return: two arrays, one with training and one with validation losses.
        """
        self.model.load_state_dict(
            torch.load(self.data_dir / f"{self.name}_weights.pth")
        )
        with open(self.data_dir / f"{self.name}_train_val_loss.pkl", "rb") as file:
            return pkl.load(file)


def plot_dataset(
    train_ds: Tuple["NDArray[np.float_]", "NDArray[np.int_]"],
    test_ds: Tuple["NDArray[np.float_]", "NDArray[np.int_]"],
    x_min: Optional["NDArray[np.float_]"] = None,
    x_max: Optional["NDArray[np.float_]"] = None,
    *,
    xlabel: Optional[str] = None,
    ylabel: Optional[str] = None,
    legend_title: Optional[str] = None,
    vline: Optional[float] = None,
    line: Optional["NDArray[np.float_]"] = None,
    suptitle: Optional[str] = None,
    s: Optional[float] = None,
    figsize: Tuple[int, int] = (20, 10),
):
    """Plots training and test data in two separate plots, with the optimal
    decision boundary as passed to the line argument.

    :param train_ds: A 2-element tuple with training data and labels thereof.
        Features have shape `(N, 2)` and the target_variable has shape `(n,)`.
    :param test_ds: A 2-element tuple with test data and labels. Same format as
        train_ds.
    :param x_min: Set to define the minimum boundaries of the plot.
    :param x_max: Set to define the maximum boundaries of the plot.
    :param line: Optional, line of shape (M,2), where each row is a point of the
        2-d line.
    :param s: The thickness of the points to plot.
    :param figsize: for `plt.figure()`
    """

    fig = plt.figure(figsize=figsize, constrained_layout=True)
    spec = fig.add_gridspec(20, 2)
    ax = [fig.add_subplot(spec[:-1, i]) for i in range(2)]
    ax.append(fig.add_subplot(spec[-1, :]))

    datasets = {"train": train_ds, "test": test_ds}

    discrete_keys = [
        key for key, dataset in datasets.items() if dataset[1].dtype == int
    ]
    if 0 < len(discrete_keys) < len(datasets):
        "You can only plot either discrete or only continuous plots."

    cmap = plt.get_cmap("Set1")
    all_y = np.concatenate(tuple([v[1] for _, v in datasets.items()]), axis=0)
    unique_y = np.sort(np.unique(all_y))
    num_classes = len(unique_y)
    handles = [mpatches.Patch(color=cmap(i), label=y) for i, y in enumerate(unique_y)]

    for i, dataset_name in enumerate(datasets.keys()):
        x, y = datasets[dataset_name]
        if x.shape[1] != 2:
            raise AttributeError("The maximum number of allowed features is 2.")

        ax[i].set_title(dataset_name)
        if x_min is not None:
            ax[i].set_xlim(x_min[0], x_max[0])  # type: ignore
        if x_max is not None:
            ax[i].set_ylim(x_min[1], x_max[1])  # type: ignore

        if line is not None:
            ax[i].plot(line[:, 0], line[:, 1], color="black")

        ax[i].scatter(x[:, 0], x[:, 1], c=cmap(y), s=s, edgecolors="black")

        if xlabel is not None:
            ax[i].set_xlabel(xlabel)
        if ylabel is not None:
            ax[i].set_ylabel(ylabel)

        if vline is not None:
            ax[i].axvline(vline, color="black", linestyle="--")

    ax[-1].legend(handles=handles, loc="center", ncol=num_classes, title=legend_title)
    ax[-1].axis("off")

    if suptitle is not None:
        plt.suptitle(suptitle)

    plt.show()


def plot_influences(
    x: "NDArray[np.float_]",
    influences: "NDArray[np.float_]",
    corrupted_indices: Optional[List[int]] = None,
    *,
    ax: Optional[plt.Axes] = None,
    xlabel: Optional[str] = None,
    ylabel: Optional[str] = None,
    legend_title: Optional[str] = None,
    line: Optional["NDArray[np.float_]"] = None,
    suptitle: Optional[str] = None,
    colorbar_limits: Optional[Tuple] = None,
) -> plt.Axes:
    """Plots the influence values of the training data with a color map.

    :param x: Input to the model, of shape (N,2) with N being the total number
        of points.
    :param influences: an array  of shape (N,) with influence values for each
        data point.
    :param line: Optional, line of shape [Mx2], where each row is a point of the
        2-dimensional line. (??)
    """
    if ax is None:
        _, ax = plt.subplots()
    sc = ax.scatter(x[:, 0], x[:, 1], c=influences)
    if line is not None:
        ax.plot(line[:, 0], line[:, 1], color="black")
    if xlabel is not None:
        ax.set_xlabel(xlabel)
    if ylabel is not None:
        ax.set_ylabel(ylabel)
    if suptitle is not None:
        plt.suptitle(suptitle)

    plt.colorbar(sc, label=legend_title)
    if colorbar_limits is not None:
        sc.clim(*colorbar_limits)
    if corrupted_indices is not None:
        ax.scatter(
            x[:, 0][corrupted_indices],
            x[:, 1][corrupted_indices],
            facecolors="none",
            edgecolors="r",
            s=80,
        )
    return ax


def plot_iris(
    data: Dataset,
    indices: List[int] = None,
    highlight_indices: Optional[Sequence[int]] = None,
    suptitle: str = None,
    legend_title: str = None,
    legend_labels: Sequence[str] = None,
    colors: Iterable = None,
    colorbar_limits: Optional[Tuple] = None,
    figsize: Tuple[int, int] = (20, 8),
):
    """Scatter plots for the iris dataset.

    :param data: a Dataset with a valid train / test split
    :param indices: subset of `data.indices`.
    :param highlight_indices: circle these indices in red
    :param suptitle: centered title for the figure
    :param legend_title: A title for the legend
    :param legend_labels: Labels for the legend entries
    :param colors: use with indices to set the color (e.g. to values).
    :param colorbar_limits: Range of values to display in the colorbar. A
        colorbar will only be displayed if there are more than 10 colors.
    :param figsize: Size of figure for matplotlib
    """
    if indices is not None:
        x_train = data.x_train[indices]
        y_train = data.y_train[indices]
    else:
        x_train = data.x_train
        y_train = data.y_train

    sepal_length_indices = data.feature("sepal length (cm)")
    sepal_width_indices = data.feature("sepal width (cm)")
    petal_length_indices = data.feature("petal length (cm)")
    petal_width_indices = data.feature("petal width (cm)")

    if colors is None:
        colors = y_train
    marker_size = 2 * plt.rcParams["lines.markersize"] ** 2

    def _handle_legend(scatter):
        if len(np.unique(colors)) > 10:
            plt.colorbar(label=legend_title)
            if colorbar_limits is not None:
                plt.clim(*colorbar_limits)
        else:
            plt.legend(
                handles=scatter.legend_elements()[0],
                labels=legend_labels,
                title=legend_title,
            )

    plt.figure(figsize=figsize)
    plt.suptitle(suptitle)
    plt.subplot(1, 2, 1)
    xmin, xmax = (
        x_train[sepal_length_indices].min(),
        x_train[sepal_length_indices].max(),
    )
    ymin, ymax = (
        x_train[sepal_width_indices].min(),
        x_train[sepal_width_indices].max(),
    )
    xmargin = 0.1 * (xmax - xmin)
    ymargin = 0.1 * (ymax - ymin)
    scatter = plt.scatter(
        x_train[sepal_length_indices],
        x_train[sepal_width_indices],
        c=colors,
        s=marker_size,
        marker="o",
        alpha=0.8,
    )
    plt.xlim(xmin - xmargin, xmax + xmargin)
    plt.ylim(ymin - ymargin, ymax + ymargin)
    plt.xlabel("Sepal length")
    plt.ylabel("Sepal width")
    _handle_legend(scatter)
    if highlight_indices is not None:
        scatter = plt.scatter(
            x_train[sepal_length_indices][highlight_indices],
            x_train[sepal_width_indices][highlight_indices],
            facecolors="none",
            edgecolors="r",
            s=marker_size * 1.1,
        )

    plt.subplot(1, 2, 2)
    xmin, xmax = (
        x_train[petal_length_indices].min(),
        x_train[petal_length_indices].max(),
    )
    ymin, ymax = (
        x_train[petal_width_indices].min(),
        x_train[petal_width_indices].max(),
    )
    xmargin = 0.1 * (xmax - xmin)
    ymargin = 0.1 * (ymax - ymin)
    scatter = plt.scatter(
        x_train[petal_length_indices],
        x_train[petal_width_indices],
        c=colors,
        marker="o",
        s=marker_size,
        alpha=0.8,
    )
    plt.xlim(xmin - xmargin, xmax + xmargin)
    plt.ylim(ymin - ymargin, ymax + ymargin)
    plt.xlabel("Petal length")
    plt.ylabel("Petal width")
    _handle_legend(scatter)
    if highlight_indices is not None:
        scatter = plt.scatter(
            x_train[petal_length_indices][highlight_indices],
            x_train[petal_width_indices][highlight_indices],
            facecolors="none",
            edgecolors="r",
            s=80,
        )


def load_preprocess_imagenet(
    train_size: float,
    test_size: float,
    downsampling_ratio: float = 1,
    keep_labels: Optional[dict] = None,
    random_state: Optional[int] = None,
) -> Tuple[pd.DataFrame, pd.DataFrame, pd.DataFrame]:
    """Loads the tiny imagenet dataset from huggingface and preprocesses it
    for model input.

    :param train_size: fraction of indices to use for training
    :param test_size: fraction of data to use for testing
    :param downsampling_ratio: which fraction of the full dataset to keep.
        E.g. downsample_ds_to_fraction=0.2 only 20% of the dataset is kept
    :param keep_labels: which of the original labels to keep and their names.
        E.g. keep_labels={10:"a", 20: "b"} only returns the images with labels
        10 and 20 and changes the values to "a" and "b" respectively.
    :param random_state: Random state. Fix this for reproducibility of sampling.
    :return: a tuple of three dataframes, first holding the training data,
    second validation, third test.
        Each has 3 keys: normalized_images has all the input images, rescaled
        to mean 0.5 and std 0.225,
        labels has the labels of each image, while images has the unmodified
        PIL images.
    """
    try:
        from datasets import load_dataset, utils
        from torchvision import transforms
    except ImportError as e:
        raise RuntimeError(
            "Torchvision and Huggingface datasets are required to load and "
            "process the imagenet dataset."
        ) from e

    utils.logging.set_verbosity_error()

    preprocess_rgb = transforms.Compose(
        [
            transforms.ToTensor(),
            transforms.Normalize(mean=[0.5, 0.5, 0.5], std=[0.225, 0.225, 0.225]),
        ]
    )

    def _process_dataset(ds):
        processed_ds = {"normalized_images": [], "labels": [], "images": []}
        for i, item in enumerate(ds):
            if item["image"].mode == "RGB":
                processed_ds["normalized_images"].append(preprocess_rgb(item["image"]))
                processed_ds["images"].append(item["image"])
                if keep_labels is not None:
                    processed_ds["labels"].append(keep_labels[item["label"]])
                else:
                    processed_ds["labels"].append(item["label"])
        return pd.DataFrame.from_dict(processed_ds)

    if os.environ.get("CI"):
        tiny_imagenet = load_dataset("Maysee/tiny-imagenet", split="valid")
        if keep_labels is not None:
            tiny_imagenet = tiny_imagenet.filter(
                lambda item: item["label"] in keep_labels
            )
        split = tiny_imagenet.shard(2, 0)
        tiny_imagenet_test = tiny_imagenet.shard(2, 1)
        tiny_imagenet_train = split.shard(5, 0)
        tiny_imagenet_val = split.shard(5, 1)
        train_ds = _process_dataset(tiny_imagenet_train)
        val_ds = _process_dataset(tiny_imagenet_val)
        test_ds = _process_dataset(tiny_imagenet_test)
        return train_ds, val_ds, test_ds

    tiny_imagenet = load_dataset("Maysee/tiny-imagenet", split="train")

    if downsampling_ratio != 1:
        tiny_imagenet = tiny_imagenet.shard(
            num_shards=int(1 / downsampling_ratio), index=0
        )
    if keep_labels is not None:
        tiny_imagenet = tiny_imagenet.filter(
            lambda item: item["label"] in keep_labels.keys()
        )

    split_ds = tiny_imagenet.train_test_split(
        train_size=1 - test_size, seed=random_state
    )
    test_ds = _process_dataset(split_ds["test"])

    split_ds = split_ds["train"].train_test_split(
        train_size=train_size, seed=random_state
    )
    train_ds = _process_dataset(split_ds["train"])
    val_ds = _process_dataset(split_ds["test"])

    return train_ds, val_ds, test_ds


def plot_sample_images(dataset: pd.DataFrame, n_images_per_class: int = 3):
    """Plots several images for each class of a pre-processed imagenet dataset
    (or a subset of it).

    :param dataset: imagenet dataset
    :param n_images_per_class: number of images per class to plot
    """
    labels = dataset["labels"].unique()
    fig, axes = plt.subplots(nrows=n_images_per_class, ncols=len(labels))
    fig.suptitle("Examples of training images")
    for class_idx, class_label in enumerate(labels):
        for img_idx, (_, img_data) in enumerate(
            dataset[dataset["labels"] == class_label].iterrows()
        ):
            axes[img_idx, class_idx].imshow(img_data["images"])
            axes[img_idx, class_idx].axis("off")
            axes[img_idx, class_idx].set_title(f"img label: {class_label}")
            if img_idx + 1 >= n_images_per_class:
                break
    plt.show()


def plot_lowest_highest_influence_images(
    subset_influences: "NDArray[np.float_]",
    subset_images: List[JpegImageFile],
    num_to_plot: int,
):
    """Given a set of images and their influences over another, plots two columns
    of `num_to_plot` images each. Those on the right column have the lowest influence,
     those on the right the highest.

    :param subset_influences: an array with influence values
    :param subset_images: a list of images
    :param num_to_plot: int, number of high and low influence images to plot
    """
    top_if_idxs = np.argsort(subset_influences)[-num_to_plot:]
    bottom_if_idxs = np.argsort(subset_influences)[:num_to_plot]

    fig, axes = plt.subplots(nrows=num_to_plot, ncols=2)
    fig.suptitle("Lowest (left) and highest (right) influences")

    for plt_idx, img_idx in enumerate(bottom_if_idxs):
        axes[plt_idx, 0].set_title(f"img influence: {subset_influences[img_idx]:0f}")
        axes[plt_idx, 0].imshow(subset_images[img_idx])
        axes[plt_idx, 0].axis("off")

    for plt_idx, img_idx in enumerate(top_if_idxs):
        axes[plt_idx, 1].set_title(f"img influence: {subset_influences[img_idx]:0f}")
        axes[plt_idx, 1].imshow(subset_images[img_idx])
        axes[plt_idx, 1].axis("off")

    plt.show()


def plot_losses(
    training_loss: "NDArray[np.float_]", validation_loss: "NDArray[np.float_]"
):
    """Plots the train and validation loss

    :param training_loss: list of training losses, one per epoch
    :param validation_loss: list of validation losses, one per epoch
    """
    _, ax = plt.subplots()
    ax.plot(training_loss, label="Train")
    ax.plot(validation_loss, label="Val")
    ax.set_ylabel("Loss")
    ax.set_xlabel("Train epoch")
    ax.legend()
    plt.show()


def corrupt_imagenet(
    dataset: pd.DataFrame,
    fraction_to_corrupt: float,
    avg_influences: "NDArray[np.float_]",
) -> Tuple[pd.DataFrame, Dict[Any, List[int]]]:
    """Given the preprocessed tiny imagenet dataset (or a subset of it), 
    it takes a fraction of the images with the highest influence and (randomly)
    flips their labels.

    :param dataset: preprocessed tiny imagenet dataset
    :param fraction_to_corrupt: float, fraction of data to corrupt
    :param avg_influences: average influences of each training point on the test set in the \
        non-corrupted case.
    :return: first element is the corrupted dataset, second is the list of indices \
        related to the images that have been corrupted.
    """
    labels = dataset["labels"].unique()
    corrupted_dataset = deepcopy(dataset)
    corrupted_indices = {l: [] for l in labels}

    avg_influences_series = pd.DataFrame()
    avg_influences_series["avg_influences"] = avg_influences
    avg_influences_series["labels"] = dataset["labels"]

    for label in labels:
        class_data = avg_influences_series[avg_influences_series["labels"] == label]
        num_corrupt = int(fraction_to_corrupt * len(class_data))
        indices_to_corrupt = class_data.nlargest(
            num_corrupt, "avg_influences"
        ).index.tolist()
        wrong_labels = [l for l in labels if l != label]
        for img_idx in indices_to_corrupt:
            sample_label = np.random.choice(wrong_labels)
            corrupted_dataset.at[img_idx, "labels"] = sample_label
            corrupted_indices[sample_label].append(img_idx)
    return corrupted_dataset, corrupted_indices


def compute_mean_corrupted_influences(
    corrupted_dataset: pd.DataFrame,
    corrupted_indices: Dict[Any, List[int]],
    avg_corrupted_influences: "NDArray[np.float_]",
) -> pd.DataFrame:
    """Given a corrupted dataset, it returns a dataframe with average influence for each class
    and separately for corrupted (and non) point.

    :param corrupted_dataset: corrupted dataset as returned by get_corrupted_imagenet
    :param corrupted_indices: list of corrupted indices, as returned by get_corrupted_imagenet
    :param avg_corrupted_influences: average influence of each training point on the test dataset
    :return: a dataframe holding the average influence of corrupted and non-corrupted data
    """
    labels = corrupted_dataset["labels"].unique()
    avg_label_influence = pd.DataFrame(
        columns=["label", "avg_non_corrupted_infl", "avg_corrupted_infl", "score_diff"]
    )
    for idx, label in enumerate(labels):
        avg_influences_series = pd.Series(avg_corrupted_influences)
        class_influences = avg_influences_series[corrupted_dataset["labels"] == label]
        corrupted_infl = class_influences[
            class_influences.index.isin(corrupted_indices[label])
        ]
        non_corrupted_infl = class_influences[
            ~class_influences.index.isin(corrupted_indices[label])
        ]
        avg_non_corrupted = np.mean(non_corrupted_infl)
        avg_corrupted = np.mean(corrupted_infl)
        avg_label_influence.loc[idx] = [
            label,
            avg_non_corrupted,
            avg_corrupted,
            avg_non_corrupted - avg_corrupted,
        ]
    return avg_label_influence


def plot_corrupted_influences_distribution(
    corrupted_dataset: pd.DataFrame,
    corrupted_indices: Dict[Any, List[int]],
    avg_corrupted_influences: "NDArray[np.float_]",
    figsize: Tuple[int, int] = (16, 8),
):
    """Given a corrupted dataset, plots the histogram with the distribution of
    influence values. This is done separately for each label: each has a plot
    where the distribution of the influence of non-corrupted points is compared
    to that of corrupted ones.

    :param corrupted_dataset: corrupted dataset as returned by
        get_corrupted_imagenet
    :param corrupted_indices: list of corrupted indices, as returned by
        get_corrupted_imagenet
    :param avg_corrupted_influences: average influence of each training point on
        the test dataset
    :param figsize: for `plt.subplots()`
    :return: a dataframe holding the average influence of corrupted and
        non-corrupted data
    """
    labels = corrupted_dataset["labels"].unique()
    fig, axes = plt.subplots(nrows=1, ncols=2, figsize=figsize)
    fig.suptitle("Distribution of corrupted and clean influences.")
    for idx, label in enumerate(labels):
        avg_influences_series = pd.Series(avg_corrupted_influences)
        class_influences = avg_influences_series[corrupted_dataset["labels"] == label]
        corrupted_infl = class_influences[
            class_influences.index.isin(corrupted_indices[label])
        ]
        non_corrupted_infl = class_influences[
            ~class_influences.index.isin(corrupted_indices[label])
        ]
        axes[idx].hist(
            non_corrupted_infl, label="Non corrupted", density=True, alpha=0.7
        )
        axes[idx].hist(
            corrupted_infl, label="Corrupted", density=True, alpha=0.7, color="green"
        )
        axes[idx].set_xlabel("Influence values")
        axes[idx].set_ylabel("Distribution")
        axes[idx].set_title(f"Influences for {label=}")
        axes[idx].legend()
    plt.show()



import logging

import numpy as np
import pytest
from sklearn.linear_model import LinearRegression

from pydvl.utils import GroupedDataset, MemcachedConfig, Utility
from pydvl.value.shapley.naive import (
    combinatorial_exact_shapley,
    permutation_exact_shapley,
)

from .. import check_total_value, check_values

log = logging.getLogger(__name__)


# noinspection PyTestParametrized
@pytest.mark.parametrize(
    "num_samples, fun, rtol, total_atol",
    [
        (12, combinatorial_exact_shapley, 0.01, 1e-5),
        (6, permutation_exact_shapley, 0.01, 1e-5),
    ],
)
def test_analytic_exact_shapley(num_samples, analytic_shapley, fun, rtol, total_atol):
    """Compares the combinatorial exact shapley and permutation exact shapley with
    the analytic_shapley calculation for a dummy model.
    """
    u, exact_values = analytic_shapley
    values_p = fun(u, progress=False)
    check_total_value(u, values_p, atol=total_atol)
    check_values(values_p, exact_values, rtol=rtol)


@pytest.mark.parametrize(
    "a, b, num_points, scorer",
    [
        (2, 0, 20, "r2"),
        (2, 1, 20, "r2"),
        (2, 1, 20, "neg_median_absolute_error"),
        (2, 1, 20, "explained_variance"),
    ],
)
def test_linear(
    linear_dataset, memcache_client_config, scorer, rtol=0.01, total_atol=1e-5
):
    linear_utility = Utility(
        LinearRegression(),
        data=linear_dataset,
        scorer=scorer,
        cache_options=MemcachedConfig(client_config=memcache_client_config),
    )

    values_combinatorial = combinatorial_exact_shapley(linear_utility, progress=False)
    check_total_value(linear_utility, values_combinatorial, atol=total_atol)

    values_permutation = permutation_exact_shapley(linear_utility, progress=False)
    check_total_value(linear_utility, values_permutation, atol=total_atol)

    check_values(values_combinatorial, values_permutation, rtol=rtol)


@pytest.mark.parametrize(
    "a, b, num_points, num_groups, scorer",
    [(2, 0, 50, 3, "r2"), (2, 1, 100, 5, "r2"), (2, 1, 100, 5, "explained_variance")],
)
def test_grouped_linear(
    linear_dataset,
    num_groups,
    memcache_client_config,
    scorer,
    rtol=0.01,
    total_atol=1e-5,
):
    # assign groups recursively
    data_groups = np.random.randint(0, num_groups, len(linear_dataset))

    grouped_linear_dataset = GroupedDataset.from_dataset(linear_dataset, data_groups)
    grouped_linear_utility = Utility(
        LinearRegression(),
        data=grouped_linear_dataset,
        scorer=scorer,
        cache_options=MemcachedConfig(client_config=memcache_client_config),
    )
    values_combinatorial = combinatorial_exact_shapley(
        grouped_linear_utility, progress=False
    )
    check_total_value(grouped_linear_utility, values_combinatorial, atol=total_atol)

    values_permutation = permutation_exact_shapley(
        grouped_linear_utility, progress=False
    )
    check_total_value(grouped_linear_utility, values_permutation, atol=total_atol)

    check_values(values_combinatorial, values_permutation, rtol=rtol)


@pytest.mark.parametrize(
    "a, b, num_points, scorer",
    [
        (2, 1, 20, "explained_variance"),
        (2, 0, 20, "r2"),
        (2, 1, 20, "neg_median_absolute_error"),
        (2, 1, 20, "r2"),
    ],
)
def test_linear_with_outlier(
    linear_dataset, memcache_client_config, scorer, total_atol=1e-5
):
    outlier_idx = np.random.randint(len(linear_dataset.y_train))
    linear_dataset.y_train[outlier_idx] -= 100
    linear_utility = Utility(
        LinearRegression(),
        data=linear_dataset,
        scorer=scorer,
        cache_options=MemcachedConfig(client_config=memcache_client_config),
    )
    values = permutation_exact_shapley(linear_utility, progress=False)
    values.sort()
    check_total_value(linear_utility, values, atol=total_atol)

    assert values.indices[0] == outlier_idx


@pytest.mark.parametrize(
    "coefficients, scorer",
    [
        (np.random.randint(-3, 3, size=3), "r2"),
        (np.random.randint(-3, 3, size=5), "neg_median_absolute_error"),
        (np.random.randint(-3, 3, size=7), "explained_variance"),
    ],
)
def test_polynomial(
    polynomial_dataset,
    polynomial_pipeline,
    memcache_client_config,
    scorer,
    rtol=0.01,
    total_atol=1e-5,
):
    dataset, _ = polynomial_dataset
    poly_utility = Utility(
        polynomial_pipeline,
        dataset,
        scorer=scorer,
        cache_options=MemcachedConfig(client_config=memcache_client_config),
    )

    values_combinatorial = combinatorial_exact_shapley(poly_utility, progress=False)
    check_total_value(poly_utility, values_combinatorial, atol=total_atol)

    values_permutation = permutation_exact_shapley(poly_utility, progress=False)
    check_total_value(poly_utility, values_permutation, atol=total_atol)

    check_values(values_combinatorial, values_permutation, rtol=rtol)


@pytest.mark.parametrize(
    "coefficients, scorer",
    [
        (np.random.randint(-3, 3, size=3), "r2"),
        (np.random.randint(-3, 3, size=3), "neg_median_absolute_error"),
        (np.random.randint(-3, 3, size=3), "explained_variance"),
    ],
)
def test_polynomial_with_outlier(
    polynomial_dataset,
    polynomial_pipeline,
    memcache_client_config,
    scorer,
    total_atol=1e-5,
):
    dataset, _ = polynomial_dataset
    outlier_idx = np.random.randint(len(dataset.y_train))
    dataset.y_train[outlier_idx] *= 100
    poly_utility = Utility(
        polynomial_pipeline,
        dataset,
        scorer=scorer,
        cache_options=MemcachedConfig(client_config=memcache_client_config),
    )

    shapley_values = permutation_exact_shapley(poly_utility, progress=False)
    check_total_value(poly_utility, shapley_values, atol=total_atol)

    assert shapley_values[0].index == outlier_idx
